#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: devices_switch_routing_interfaces
short_description: Resource module for devices _switch _routing _interfaces
description:
  - Manage operations create, update and delete of the resource devices _switch _routing
    _interfaces.
  - Create a layer 3 interface for a switch.
  - Delete a layer 3 interface from the switch.
  - Update a layer 3 interface for a switch.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module
author: Francisco Munoz (@fmunoz)
options:
  defaultGateway:
    description: The next hop for any traffic that isn't going to a directly connected
      subnet or over a static route. This IP address must exist in a subnet with a
      L3 interface. Required if this is the first IPv4 interface.
    type: str
  interfaceId:
    description: InterfaceId path parameter. Interface ID.
    type: str
  interfaceIp:
    description: The IP address that will be used for Layer 3 routing on this VLAN
      or subnet. This cannot be the same as the device management IP.
    type: str
  ipv6:
    description: The IPv6 settings of the interface.
    suboptions:
      address:
        description: The IPv6 address of the interface. Required if assignmentMode
          is 'static'. Must not be included if assignmentMode is 'eui-64'.
        type: str
      assignmentMode:
        description: The IPv6 assignment mode for the interface. Can be either 'eui-64'
          or 'static'.
        type: str
      gateway:
        description: The IPv6 default gateway of the interface. Required if prefix
          is defined and this is the first interface with IPv6 configured.
        type: str
      prefix:
        description: The IPv6 prefix of the interface. Required if IPv6 object is
          included.
        type: str
    type: dict
  loopback:
    description: The loopback settings of the interface.
    type: dict
  mode:
    description: L3 Interface mode, can be one of 'vlan', 'routed', 'loopback'. Default
      is 'vlan'. CS 17.18 or higher is required for 'routed' mode.
    type: str
  multicastRouting:
    description: Enable multicast support if, multicast routing between VLANs is required.
      Options are 'disabled', 'enabled' or 'IGMP snooping querier'. Default is 'disabled'.
    type: str
  name:
    description: A friendly name or description for the interface or VLAN (max length
      128 characters).
    type: str
  ospfSettings:
    description: The OSPF routing settings of the interface.
    suboptions:
      area:
        description: The OSPF area to which this interface should belong. Can be either
          'disabled' or the identifier of an existing OSPF area. Defaults to 'disabled'.
        type: str
      cost:
        description: The path cost for this interface. Defaults to 1, but can be increased
          up to 65535 to give lower priority.
        type: int
      isPassiveEnabled:
        description: When enabled, OSPF will not run on the interface, but the subnet
          will still be advertised.
        type: bool
      networkType:
        description: OSPF network type.
        type: str
    type: dict
  serial:
    description: Serial path parameter.
    type: str
  subnet:
    description: The network that this L3 interface is on, in CIDR notation (ex. 10.1.1.0/24).
    type: str
  switchPortId:
    description: Switch Port ID when in Routed mode (CS 17.18 or higher required).
    type: str
  vlanId:
    description: The VLAN this L3 interface is on. VLAN must be between 1 and 4094.
    type: int
  vrf:
    description: The VRF settings of the interface. Requires IOS XE 17.18 or higher.
    suboptions:
      name:
        description: The name of the VRF this interface belongs to.
        type: str
    type: dict
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for switch createDeviceSwitchRoutingInterface
    description: Complete reference of the createDeviceSwitchRoutingInterface API.
    link: https://developer.cisco.com/meraki/api-v1/#!create-device-switch-routing-interface
  - name: Cisco Meraki documentation for switch deleteDeviceSwitchRoutingInterface
    description: Complete reference of the deleteDeviceSwitchRoutingInterface API.
    link: https://developer.cisco.com/meraki/api-v1/#!delete-device-switch-routing-interface
  - name: Cisco Meraki documentation for switch updateDeviceSwitchRoutingInterface
    description: Complete reference of the updateDeviceSwitchRoutingInterface API.
    link: https://developer.cisco.com/meraki/api-v1/#!update-device-switch-routing-interface
notes:
  - SDK Method used are
    switch.Switch.create_device_switch_routing_interface,
    switch.Switch.delete_device_switch_routing_interface,
    switch.Switch.update_device_switch_routing_interface,
  - Paths used are
    post /devices/{serial}/switch/routing/interfaces,
    delete /devices/{serial}/switch/routing/interfaces/{interfaceId},
    put /devices/{serial}/switch/routing/interfaces/{interfaceId},
"""

EXAMPLES = r"""
- name: Create
  cisco.meraki.devices_switch_routing_interfaces:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: present
    defaultGateway: 192.168.1.1
    interfaceIp: 192.168.1.2
    ipv6:
      address: 2001:db8::1
      assignmentMode: static
      gateway: 2001:db8::2
      prefix: 2001:db8::/32
    mode: vlan
    multicastRouting: disabled
    name: L3 interface
    ospfSettings:
      area: '0'
      cost: 1
      isPassiveEnabled: true
      networkType: broadcast
    serial: string
    subnet: 192.168.1.0/24
    switchPortId: '1'
    vlanId: 100
    vrf:
      name: Blue
- name: Update by id
  cisco.meraki.devices_switch_routing_interfaces:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: present
    defaultGateway: 192.168.1.1
    interfaceId: string
    interfaceIp: 192.168.1.2
    ipv6:
      address: 2001:db8::1
      assignmentMode: static
      gateway: 2001:db8::2
      prefix: 2001:db8::/32
    multicastRouting: disabled
    name: L3 interface
    ospfSettings:
      area: '0'
      cost: 1
      isPassiveEnabled: true
      networkType: broadcast
    serial: string
    subnet: 192.168.1.0/24
    switchPortId: '1'
    vlanId: 100
    vrf:
      name: Blue
- name: Delete by id
  cisco.meraki.devices_switch_routing_interfaces:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: absent
    interfaceId: string
    serial: string
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: dict
  sample: >
    {
      "interfaceId": "string",
      "name": "string",
      "mode": "string",
      "subnet": "string",
      "interfaceIp": "string",
      "serial": "string",
      "switchPortId": "string",
      "multicastRouting": "string",
      "vlanId": 0,
      "uplinkV4": true,
      "uplinkV6": true,
      "ospfSettings": {
        "area": "string",
        "cost": 0,
        "isPassiveEnabled": true,
        "networkType": "string"
      },
      "ospfV3": {
        "area": "string",
        "cost": 0,
        "isPassiveEnabled": true,
        "networkType": "string"
      },
      "ipv6": {
        "assignmentMode": "string",
        "address": "string",
        "prefix": "string",
        "gateway": "string"
      },
      "vrf": {
        "name": "string"
      },
      "loopback": {},
      "defaultGateway": "string"
    }
"""



#!/usr/bin/python

# Copyright (c) 2016, Alain Dejoux <adejoux@djouxtech.net>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
author:
  - Alain Dejoux (@adejoux)
module: aix_lvol
short_description: Configure AIX LVM logical volumes
description:
  - This module creates, removes or resizes AIX logical volumes. Inspired by M(community.general.lvol) module.
deprecated:
  removed_in: 15.0.0
  why: The module is not actively maintained.
  alternative: >-
    Use C(ibm.power_aix.lvol) instead.
    See U(https://ibm.github.io/ansible-power-aix/modules/lvol.html) for details.
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  vg:
    description:
      - The volume group this logical volume is part of.
    type: str
    required: true
  lv:
    description:
      - The name of the logical volume.
    type: str
    required: true
  lv_type:
    description:
      - The type of the logical volume.
    type: str
    default: jfs2
  size:
    description:
      - The size of the logical volume with one of the [MGT] units.
    type: str
  copies:
    description:
      - The number of copies of the logical volume.
      - Maximum copies are 3.
    type: int
    default: 1
  policy:
    description:
      - Sets the interphysical volume allocation policy.
      - V(maximum) allocates logical partitions across the maximum number of physical volumes.
      - V(minimum) allocates logical partitions across the minimum number of physical volumes.
    type: str
    choices: [maximum, minimum]
    default: maximum
  state:
    description:
      - Control if the logical volume exists. If V(present) and the volume does not already exist then the O(size) option
        is required.
    type: str
    choices: [absent, present]
    default: present
  opts:
    description:
      - Free-form options to be passed to the mklv command.
    type: str
    default: ''
  pvs:
    description:
      - A list of physical volumes, for example V([hdisk1, hdisk2]).
    type: list
    elements: str
    default: []
"""

EXAMPLES = r"""
- name: Create a logical volume of 512M
  community.general.aix_lvol:
    vg: testvg
    lv: testlv
    size: 512M

- name: Create a logical volume of 512M with disks hdisk1 and hdisk2
  community.general.aix_lvol:
    vg: testvg
    lv: test2lv
    size: 512M
    pvs: [hdisk1, hdisk2]

- name: Create a logical volume of 512M mirrored
  community.general.aix_lvol:
    vg: testvg
    lv: test3lv
    size: 512M
    copies: 2

- name: Create a logical volume of 1G with a minimum placement policy
  community.general.aix_lvol:
    vg: rootvg
    lv: test4lv
    size: 1G
    policy: minimum

- name: Create a logical volume with special options like mirror pool
  community.general.aix_lvol:
    vg: testvg
    lv: testlv
    size: 512M
    opts: -p copy1=poolA -p copy2=poolB

- name: Extend the logical volume to 1200M
  community.general.aix_lvol:
    vg: testvg
    lv: test4lv
    size: 1200M

- name: Remove the logical volume
  community.general.aix_lvol:
    vg: testvg
    lv: testlv
    state: absent
"""

RETURN = r"""
msg:
  type: str
  description: A friendly message describing the task result.
  returned: always
  sample: Logical volume testlv created.
"""

import re

from ansible.module_utils.basic import AnsibleModule


def convert_size(module, size):
    unit = size[-1].upper()
    units = ["M", "G", "T"]
    try:
        multiplier = 1024 ** units.index(unit)
    except ValueError:
        module.fail_json(msg="No valid size unit specified.")

    return int(size[:-1]) * multiplier


def round_ppsize(x, base=16):
    new_size = int(base * round(float(x) / base))
    if new_size < x:
        new_size += base
    return new_size


def parse_lv(data):
    name = None

    for line in data.splitlines():
        match = re.search(r"LOGICAL VOLUME:\s+(\w+)\s+VOLUME GROUP:\s+(\w+)", line)
        if match is not None:
            name = match.group(1)
            vg = match.group(2)
            continue
        match = re.search(r"LPs:\s+(\d+).*PPs", line)
        if match is not None:
            lps = int(match.group(1))
            continue
        match = re.search(r"PP SIZE:\s+(\d+)", line)
        if match is not None:
            pp_size = int(match.group(1))
            continue
        match = re.search(r"INTER-POLICY:\s+(\w+)", line)
        if match is not None:
            policy = match.group(1)
            continue

    if not name:
        return None

    size = lps * pp_size

    return {"name": name, "vg": vg, "size": size, "policy": policy}


def parse_vg(data):
    for line in data.splitlines():
        match = re.search(r"VOLUME GROUP:\s+(\w+)", line)
        if match is not None:
            name = match.group(1)
            continue

        match = re.search(r"TOTAL PP.*\((\d+)", line)
        if match is not None:
            size = int(match.group(1))
            continue

        match = re.search(r"PP SIZE:\s+(\d+)", line)
        if match is not None:
            pp_size = int(match.group(1))
            continue

        match = re.search(r"FREE PP.*\((\d+)", line)
        if match is not None:
            free = int(match.group(1))
            continue

    return {"name": name, "size": size, "free": free, "pp_size": pp_size}


def main():
    module = AnsibleModule(
        argument_spec=dict(
            vg=dict(type="str", required=True),
            lv=dict(type="str", required=True),
            lv_type=dict(type="str", default="jfs2"),
            size=dict(type="str"),
            opts=dict(type="str", default=""),
            copies=dict(type="int", default=1),
            state=dict(type="str", default="present", choices=["absent", "present"]),
            policy=dict(type="str", default="maximum", choices=["maximum", "minimum"]),
            pvs=dict(type="list", elements="str", default=list()),
        ),
        supports_check_mode=True,
    )

    vg = module.params["vg"]
    lv = module.params["lv"]
    lv_type = module.params["lv_type"]
    size = module.params["size"]
    opts = module.params["opts"]
    copies = module.params["copies"]
    policy = module.params["policy"]
    state = module.params["state"]
    pvs = module.params["pvs"]

    if policy == "maximum":
        lv_policy = "x"
    else:
        lv_policy = "m"

    # Add echo command when running in check-mode
    if module.check_mode:
        test_opt = [module.get_bin_path("echo", required=True)]
    else:
        test_opt = []

    # check if system commands are available
    lsvg_cmd = module.get_bin_path("lsvg", required=True)
    lslv_cmd = module.get_bin_path("lslv", required=True)

    # Get information on volume group requested
    rc, vg_info, err = module.run_command([lsvg_cmd, vg])

    if rc != 0:
        if state == "absent":
            module.exit_json(changed=False, msg=f"Volume group {vg} does not exist.")
        else:
            module.fail_json(msg=f"Volume group {vg} does not exist.", rc=rc, out=vg_info, err=err)

    this_vg = parse_vg(vg_info)

    if size is not None:
        # Calculate pp size and round it up based on pp size.
        lv_size = round_ppsize(convert_size(module, size), base=this_vg["pp_size"])

    # Get information on logical volume requested
    rc, lv_info, err = module.run_command([lslv_cmd, lv])

    if rc != 0:
        if state == "absent":
            module.exit_json(changed=False, msg=f"Logical Volume {lv} does not exist.")

    this_lv = parse_lv(lv_info)

    if state == "present" and not size:
        if this_lv is None:
            module.fail_json(msg="No size given.")

    if this_lv is None:
        if state == "present":
            if lv_size > this_vg["free"]:
                module.fail_json(
                    msg=f"Not enough free space in volume group {this_vg['name']}: {this_vg['free']} MB free."
                )

            # create LV
            mklv_cmd = module.get_bin_path("mklv", required=True)

            cmd = (
                test_opt
                + [mklv_cmd, "-t", lv_type, "-y", lv, "-c", copies, "-e", lv_policy, opts, vg, f"{lv_size}M"]
                + pvs
            )
            rc, out, err = module.run_command(cmd)
            if rc == 0:
                module.exit_json(changed=True, msg=f"Logical volume {lv} created.")
            else:
                module.fail_json(msg=f"Creating logical volume {lv} failed.", rc=rc, out=out, err=err)
    else:
        if state == "absent":
            # remove LV
            rmlv_cmd = module.get_bin_path("rmlv", required=True)
            rc, out, err = module.run_command(test_opt + [rmlv_cmd, "-f", this_lv["name"]])
            if rc == 0:
                module.exit_json(changed=True, msg=f"Logical volume {lv} deleted.")
            else:
                module.fail_json(msg=f"Failed to remove logical volume {lv}.", rc=rc, out=out, err=err)
        else:
            if this_lv["policy"] != policy:
                # change lv allocation policy
                chlv_cmd = module.get_bin_path("chlv", required=True)
                rc, out, err = module.run_command(test_opt + [chlv_cmd, "-e", lv_policy, this_lv["name"]])
                if rc == 0:
                    module.exit_json(changed=True, msg=f"Logical volume {lv} policy changed: {policy}.")
                else:
                    module.fail_json(msg=f"Failed to change logical volume {lv} policy.", rc=rc, out=out, err=err)

            if vg != this_lv["vg"]:
                module.fail_json(msg=f"Logical volume {lv} already exist in volume group {this_lv['vg']}")

            # from here the last remaining action is to resize it, if no size parameter is passed we do nothing.
            if not size:
                module.exit_json(changed=False, msg=f"Logical volume {lv} already exist.")

            # resize LV based on absolute values
            if int(lv_size) > this_lv["size"]:
                extendlv_cmd = module.get_bin_path("extendlv", required=True)
                cmd = test_opt + [extendlv_cmd, lv, f"{lv_size - this_lv['size']}M"]
                rc, out, err = module.run_command(cmd)
                if rc == 0:
                    module.exit_json(changed=True, msg=f"Logical volume {lv} size extended to {lv_size}MB.")
                else:
                    module.fail_json(msg=f"Unable to resize {lv} to {lv_size}MB.", rc=rc, out=out, err=err)
            elif lv_size < this_lv["size"]:
                module.fail_json(
                    msg=f"No shrinking of Logical Volume {lv} permitted. Current size: {this_lv['size']} MB"
                )
            else:
                module.exit_json(changed=False, msg=f"Logical volume {lv} size is already {lv_size}MB.")


if __name__ == "__main__":
    main()



# Copyright (c) 2017-2021 Felix Fontein
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

# The API documentation can be found here: https://api.ns1.hosttech.eu/api/documentation/

from __future__ import annotations

import typing as t

from ansible_collections.community.dns.plugins.module_utils._json_api_helper import (
    JSONAPIHelper,
)
from ansible_collections.community.dns.plugins.module_utils._record import DNSRecord
from ansible_collections.community.dns.plugins.module_utils._zone import (
    DNSZone,
    DNSZoneWithRecords,
)
from ansible_collections.community.dns.plugins.module_utils._zone_record_api import (
    NOT_PROVIDED,
    DNSAPIError,
    ZoneRecordAPI,
    filter_records,
)

if t.TYPE_CHECKING:
    from collections.abc import Sequence  # pragma: no cover

    from .._http import HTTPHelper  # pragma: no cover
    from .._record import IDNSRecord  # pragma: no cover
    from .._zone_record_api import NotProvidedType  # pragma: no cover


def _create_record_from_json(
    source: dict[str, t.Any], record_type: str | None = None
) -> DNSRecord[int]:
    source = dict(source)
    record_type = source.pop("type", record_type)
    if record_type is None:
        raise DNSAPIError("Record from API has no type")

    name: str | None = source.pop("name", None)
    target: str
    if record_type == "A":
        target = source.pop("ipv4")
    elif record_type == "AAAA":
        target = source.pop("ipv6")
    elif record_type == "CAA":
        target = f"{source.pop('flag')} {source.pop('tag')} \"{source.pop('value')}\""
    elif record_type == "CNAME":
        target = source.pop("cname")
    elif record_type == "MX":
        mx_name, name = name, source.pop("ownername")
        target = f"{source.pop('pref')} {mx_name}"
    elif record_type == "NS":
        name = source.pop("ownername")
        target = source.pop("targetname")
    elif record_type == "PTR":
        ptr_name, name = name, ""
        target = f"{source.pop('origin')} {ptr_name}"
    elif record_type == "SRV":
        name = source.pop("service")
        target = f"{source.pop('priority')} {source.pop('weight')} {source.pop('port')} {source.pop('target')}"
    elif record_type in ("TXT", "TLSA"):
        target = source.pop("text")
    else:
        raise DNSAPIError(f"Cannot parse unknown record type: {record_type}")

    result = DNSRecord(
        record_id=source.pop("id"), record_type=record_type, target=target
    )
    ttl = source.pop("ttl")
    result.ttl = int(ttl) if ttl is not None else None
    result.extra["comment"] = source.pop("comment")
    result.prefix = name or None  # API returns '', we want None
    result.extra.update(source)
    return result


def _create_zone_from_json(source: dict[str, t.Any]) -> DNSZone[int]:
    zone = DNSZone(zone_id=source["id"], name=source["name"])
    zone.info = {
        "dnssec": source["dnssec"],
        "dnssec_email": source.get("dnssec_email"),
        "ds_records": source.get("ds_records"),
        "email": source.get("email"),
        "ttl": source["ttl"],
    }
    return zone


def _create_zone_with_records_from_json(
    source: dict[str, t.Any],
    prefix: str | None | NotProvidedType = NOT_PROVIDED,
    record_type: str | NotProvidedType = NOT_PROVIDED,
) -> DNSZoneWithRecords[int, int]:
    return DNSZoneWithRecords(
        _create_zone_from_json(source),
        filter_records(
            [_create_record_from_json(record) for record in source["records"]],
            prefix=prefix,
            record_type=record_type,
        ),
    )


def _record_to_json(
    record: IDNSRecord[int | None], include_id: bool = False, include_type: bool = True
) -> dict[str, t.Any]:
    result: dict[str, t.Any] = {
        "ttl": record.ttl,
        "comment": record.extra.get("comment") or "",
    }
    if include_type:
        result["type"] = record.type
    if include_id:
        result["id"] = record.id

    if record.type == "A":
        result["name"] = record.prefix or ""
        result["ipv4"] = record.target
    elif record.type == "AAAA":
        result["name"] = record.prefix or ""
        result["ipv6"] = record.target
    elif record.type == "CAA":
        result["name"] = record.prefix or ""
        try:
            flag, tag, value = record.target.split(" ", 2)
            if value.startswith('"') and value.endswith('"'):
                value = value[1:-1]
            result["flag"] = flag
            result["tag"] = tag
            result["value"] = value
        except Exception as e:
            raise DNSAPIError(
                f'Cannot split {record.type} record "{record.target}" into flag, tag and value: {e}'
            ) from e
    elif record.type == "CNAME":
        result["name"] = record.prefix or ""
        result["cname"] = record.target
    elif record.type == "MX":
        result["ownername"] = record.prefix or ""
        try:
            pref, name = record.target.split(" ", 1)
            result["pref"] = int(pref)
            result["name"] = name
        except Exception as e:
            raise DNSAPIError(
                f'Cannot split {record.type} record "{record.target}" into integer preference and name: {e}'
            ) from e
    elif record.type == "NS":
        result["ownername"] = record.prefix or ""
        result["targetname"] = record.target
    elif record.type == "PTR":
        try:
            origin, name = record.target.split(" ", 1)
            result["origin"] = origin
            result["name"] = name
        except Exception as e:
            raise DNSAPIError(
                f'Cannot split {record.type} record "{record.target}" into origin and name: {e}'
            ) from e
    elif record.type == "SRV":
        result["service"] = record.prefix or ""
        try:
            priority, weight, port, target = record.target.split(" ", 3)
            result["priority"] = int(priority)
            result["weight"] = int(weight)
            result["port"] = int(port)
            result["target"] = target
        except Exception as e:
            raise DNSAPIError(
                f'Cannot split {record.type} record "{record.target}" into integer priority, integer weight, integer port and target: {e}'
            ) from e
    elif record.type in ("TXT", "TLSA"):
        result["name"] = record.prefix or ""
        result["text"] = record.target
    else:
        raise DNSAPIError(f"Cannot serialize unknown record type: {record.type}")

    return result


def _pagination_query(
    query: dict[str, str] | Sequence[tuple[str, str]] | None,
    *,
    block_size: int,
    offset: int,
) -> dict[str, str] | Sequence[tuple[str, str]]:
    if isinstance(query, dict):
        query_map = query.copy()
        query_map["limit"] = str(block_size)
        query_map["offset"] = str(offset)
        return query_map
    query_list = list(query) if query else []
    query_list.append(("limit", str(block_size)))
    query_list.append(("offset", str(offset)))
    return query_list


class HostTechJSONAPI(ZoneRecordAPI[int, int], JSONAPIHelper):
    def __init__(
        self,
        http_helper: HTTPHelper,
        token: str,
        api="https://api.ns1.hosttech.eu/api/",
        debug: bool = False,
    ) -> None:
        """
        Create a new HostTech API instance with given API token.
        """
        JSONAPIHelper.__init__(self, http_helper, token, api=api, debug=debug)

    def _extract_error_message(self, result: t.Any | None) -> str:
        if result is None:
            return ""
        if isinstance(result, dict):
            res = ""
            if result.get("message"):
                res = f"{res} with message \"{result['message']}\""
            if "errors" in result and isinstance(result["errors"], dict):
                for k, v in sorted(result["errors"].items()):
                    if isinstance(v, list):
                        v = "; ".join(v)
                    res = f'{res} (field "{k}": {v})'
            if res:
                return res
        return f" with data: {result}"

    def _create_headers(self) -> dict[str, str]:
        return {
            "accept": "application/json",
            "authorization": f"Bearer {self._token}",
        }

    def _list_pagination(
        self,
        url: str,
        query: dict[str, str] | Sequence[tuple[str, str]] | None = None,
        block_size: int = 100,
    ) -> list[t.Any]:
        result = []
        offset = 0
        while True:
            query_ = _pagination_query(query, block_size=block_size, offset=offset)
            res, dummy = self._get(
                url,
                query=query_,
                must_have_content=True,
                expected=[200],
                require_json_object=True,
            )
            result.extend(res["data"])
            if len(res["data"]) < block_size:
                return result
            offset += block_size

    def get_zone_with_records_by_id(
        self,
        zone_id: int,
        prefix: str | None | NotProvidedType = NOT_PROVIDED,
        record_type: str | NotProvidedType = NOT_PROVIDED,
    ) -> DNSZoneWithRecords[int, int] | None:
        """
        Given a zone ID, return the zone contents with records if found.

        @param zone_id: The zone ID
        @param prefix: The prefix to filter for, if provided. Since None is a valid value,
                       the special constant NOT_PROVIDED indicates that we are not filtering.
        @param record_type: The record type to filter for, if provided
        @return The zone information with records (DNSZoneWithRecords), or None if not found
        """
        result, info = self._get(
            f"user/v1/zones/{zone_id}",
            expected=[200, 404],
            must_have_content=[200],
            require_json_object=True,
        )
        if info["status"] == 404 or result is None:
            return None
        return _create_zone_with_records_from_json(
            result["data"], prefix=prefix, record_type=record_type
        )

    def get_zone_with_records_by_name(
        self,
        name: str,
        prefix: str | None | NotProvidedType = NOT_PROVIDED,
        record_type: str | NotProvidedType = NOT_PROVIDED,
    ) -> DNSZoneWithRecords[int, int] | None:
        """
        Given a zone name, return the zone contents with records if found.

        @param name: The zone name (string)
        @param prefix: The prefix to filter for, if provided. Since None is a valid value,
                       the special constant NOT_PROVIDED indicates that we are not filtering.
        @param record_type: The record type to filter for, if provided
        @return The zone information with records (DNSZoneWithRecords), or None if not found
        """
        result = self._list_pagination("user/v1/zones", query={"query": name})
        for zone in result:
            if zone["name"] == name:
                zone_result, dummy = self._get(
                    f"user/v1/zones/{zone['id']}",
                    expected=[200],
                    require_json_object=True,
                )
                return _create_zone_with_records_from_json(
                    zone_result["data"], prefix=prefix, record_type=record_type
                )
        return None

    def get_zone_records(
        self,
        zone_id: int,
        prefix: str | None | NotProvidedType = NOT_PROVIDED,
        record_type: str | NotProvidedType = NOT_PROVIDED,
    ) -> list[DNSRecord[int]] | None:
        """
        Given a zone ID, return a list of records, optionally filtered by the provided criteria.

        @param zone_id: The zone ID
        @param prefix: The prefix to filter for, if provided. Since None is a valid value,
                       the special constant NOT_PROVIDED indicates that we are not filtering.
        @param record_type: The record type to filter for, if provided
        @return A list of DNSrecord objects, or None if zone was not found
        """
        query: dict[str, str] = {}
        if record_type is not NOT_PROVIDED:
            query["type"] = record_type.upper()  # type: ignore  # Is a string
        result, info = self._get(
            f"user/v1/zones/{zone_id}/records",
            query=query,
            expected=[200, 404],
            must_have_content=[200],
            require_json_object=True,
        )
        if info["status"] == 404 or result is None:
            return None
        return filter_records(
            [_create_record_from_json(record) for record in result["data"]],
            prefix=prefix,
            record_type=record_type,
        )

    def get_zone_by_name(self, name: str) -> DNSZone[int] | None:
        """
        Given a zone name, return the zone contents if found.

        @param name: The zone name (string)
        @return The zone information (DNSZone), or None if not found
        """
        result = self._list_pagination("user/v1/zones", query={"query": name})
        for zone in result:
            if zone["name"] == name:
                # We cannot simply return `_create_zone_from_json(zone)`, since this contains less information!
                return self.get_zone_by_id(zone["id"])
        return None

    def get_zone_by_id(self, zone_id: int) -> DNSZone[int] | None:
        """
        Given a zone ID, return the zone contents if found.

        @param zone_id: The zone ID
        @return The zone information (DNSZone), or None if not found
        """
        result, info = self._get(
            f"user/v1/zones/{zone_id}",
            expected=[200, 404],
            must_have_content=[200],
            require_json_object=True,
        )
        if info["status"] == 404 or result is None:
            return None
        return _create_zone_from_json(result["data"])

    def add_record(
        self, zone_id: int, record: IDNSRecord[int | None]
    ) -> DNSRecord[int]:
        """
        Adds a new record to an existing zone.

        @param zone_id: The zone ID
        @param record: The DNS record (DNSRecord)
        @return The created DNS record (DNSRecord)
        """
        data = _record_to_json(record, include_id=False, include_type=True)
        result, dummy = self._post(
            f"user/v1/zones/{zone_id}/records",
            data=data,
            expected=[201],
            require_json_object=True,
        )
        return _create_record_from_json(result["data"])

    def update_record(self, zone_id: int, record: DNSRecord[int]) -> DNSRecord[int]:
        """
        Update a record.

        @param zone_id: The zone ID
        @param record: The DNS record (DNSRecord)
        @return The DNS record (DNSRecord)
        """
        if record.id is None:
            raise DNSAPIError("Need record ID to update record!")
        data = _record_to_json(record, include_id=False, include_type=False)
        result, dummy = self._put(
            f"user/v1/zones/{zone_id}/records/{record.id}",
            data=data,
            expected=[200],
            require_json_object=True,
        )
        return _create_record_from_json(result["data"])

    def delete_record(self, zone_id: int, record: DNSRecord[int]) -> bool:
        """
        Delete a record.

        @param zone_id: The zone ID
        @param record: The DNS record (DNSRecord)
        @return True in case of success (boolean)
        """
        if record.id is None:
            raise DNSAPIError("Need record ID to delete record!")
        dummy, info = self._delete(
            f"user/v1/zones/{zone_id}/records/{record.id}",
            must_have_content=False,
            expected=[204, 404],
            require_json_object=True,
        )
        return info["status"] == 204



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: ec2_instance_info
version_added: 1.0.0
short_description: Gather information about ec2 instances in AWS
description:
    - Gather information about EC2 instances in AWS.
author:
  - Michael Schuett (@michaeljs1990)
  - Rob White (@wimnat)
options:
  instance_ids:
    description:
      - If you specify one or more instance IDs, only instances that have the specified IDs are returned.
    required: false
    type: list
    elements: str
    default: []
  filters:
    description:
      - A dict of filters to apply. Each dict item consists of a filter key and a filter value. See
        U(https://docs.aws.amazon.com/AWSEC2/latest/APIReference/API_DescribeInstances.html) for possible filters. Filter
        names and values are case sensitive.
    required: false
    default: {}
    type: dict
  minimum_uptime:
    description:
      - Minimum running uptime in minutes of instances.  For example if O(uptime) is V(60) return all instances that have run more than 60 minutes.
    required: false
    aliases: ['uptime']
    type: int
  include_attributes:
    description:
      - Describes the specified attributes of the returned instances.
    required: false
    type: list
    elements: str
    choices:
    - instanceType
    - kernel
    - ramdisk
    - userData
    - disableApiTermination
    - instanceInitiatedShutdownBehavior
    - rootDeviceName
    - blockDeviceMapping
    - productCodes
    - sourceDestCheck
    - groupSet
    - ebsOptimized
    - sriovNetSupport
    - enclaveOptions
    - disableApiStop
    aliases: ['attributes']
    version_added: 6.3.0

extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.boto3
"""

EXAMPLES = r"""
# Note: These examples do not set authentication details, see the AWS Guide for details.

- name: Gather information about all instances
  amazon.aws.ec2_instance_info:

- name: Gather information about all instances in AZ ap-southeast-2a
  amazon.aws.ec2_instance_info:
    filters:
      availability-zone: ap-southeast-2a

- name: Gather information about a particular instance using ID
  amazon.aws.ec2_instance_info:
    instance_ids:
      - i-12345678

- name: Gather information about any instance with a tag key Name and value Example
  amazon.aws.ec2_instance_info:
    filters:
      "tag:Name": Example

- name: Gather information about any instance in states "shutting-down", "stopping", "stopped"
  amazon.aws.ec2_instance_info:
    filters:
      instance-state-name: ["shutting-down", "stopping", "stopped"]

- name: Gather information about any instance with Name beginning with RHEL and an uptime of at least 60 minutes
  amazon.aws.ec2_instance_info:
    region: "{{ ec2_region }}"
    uptime: 60
    filters:
      "tag:Name": "RHEL-*"
      instance-state-name: ["running"]
  register: ec2_node_info

- name: Gather information about a particular instance using ID and include kernel attribute
  amazon.aws.ec2_instance_info:
    instance_ids:
      - i-12345678
    include_attributes:
      - kernel
"""

RETURN = r"""
instances:
    description: A list of ec2 instances.
    returned: always
    type: complex
    contains:
        ami_launch_index:
            description: The AMI launch index, which can be used to find this instance in the launch group.
            returned: always
            type: int
            sample: 0
        architecture:
            description: The architecture of the image.
            returned: always
            type: str
            sample: x86_64
        block_device_mappings:
            description: Any block device mapping entries for the instance.
            returned: always
            type: complex
            contains:
                device_name:
                    description: The device name exposed to the instance (for example, /dev/sdh or xvdh).
                    returned: always
                    type: str
                    sample: /dev/sdh
                ebs:
                    description: Parameters used to automatically set up EBS volumes when the instance is launched.
                    returned: always
                    type: complex
                    contains:
                        attach_time:
                            description: The time stamp when the attachment initiated.
                            returned: always
                            type: str
                            sample: "2017-03-23T22:51:24+00:00"
                        delete_on_termination:
                            description: Indicates whether the volume is deleted on instance termination.
                            returned: always
                            type: bool
                            sample: true
                        status:
                            description: The attachment state.
                            returned: always
                            type: str
                            sample: attached
                        volume_id:
                            description: The ID of the EBS volume.
                            returned: always
                            type: str
                            sample: vol-12345678
        capacity_reservation_specification:
            description: Information about the Capacity Reservation targeting option.
            type: complex
            contains:
                capacity_reservation_preference:
                    description: Describes the Capacity Reservation preferences.
                    type: str
                    sample: open
        cpu_options:
            description: The CPU options set for the instance.
            returned: always
            type: complex
            contains:
                core_count:
                     description: The number of CPU cores for the instance.
                     returned: always
                     type: int
                     sample: 1
                threads_per_core:
                     description: The number of threads per CPU core. On supported instance, a value of 1 means Intel Hyper-Threading Technology is disabled.
                     returned: always
                     type: int
                     sample: 1
        client_token:
            description: The idempotency token you provided when you launched the instance, if applicable.
            returned: always
            type: str
            sample: mytoken
        current_instance_boot_mode:
            description: The boot mode that is used to boot the instance at launch or start.
            type: str
            sample: legacy-bios
        ebs_optimized:
            description: Indicates whether the instance is optimized for EBS I/O.
            returned: always
            type: bool
            sample: false
        ena_support:
            description: Specifies whether enhanced networking with ENA is enabled.
            returned: always
            type: bool
            sample: true
        enclave_options:
            description: Indicates whether the instance is enabled for Amazon Web Services Nitro Enclaves.
            type: dict
            contains:
                enabled:
                    description: If this parameter is set to true, the instance is enabled for Amazon Web Services Nitro Enclaves.
                    returned: always
                    type: bool
                    sample: false
        hibernation_options:
            description: Indicates whether the instance is enabled for hibernation.
            type: dict
            contains:
                configured:
                    description: If true, your instance is enabled for hibernation; otherwise, it is not enabled for hibernation.
                    returned: always
                    type: bool
                    sample: false
        hypervisor:
            description: The hypervisor type of the instance.
            returned: always
            type: str
            sample: xen
        iam_instance_profile:
            description: The IAM instance profile associated with the instance, if applicable.
            type: complex
            contains:
                arn:
                    description: The Amazon Resource Name (ARN) of the instance profile.
                    returned: always
                    type: str
                    sample: "arn:aws:iam::123456789012:instance-profile/myprofile"
                id:
                    description: The ID of the instance profile.
                    returned: always
                    type: str
                    sample: JFJ397FDG400FG9FD1N
        image_id:
            description: The ID of the AMI used to launch the instance.
            returned: always
            type: str
            sample: ami-0011223344
        instance_id:
            description: The ID of the instance.
            returned: always
            type: str
            sample: i-012345678
        instance_type:
            description: The instance type size of the running instance.
            returned: always
            type: str
            sample: t2.micro
        key_name:
            description: The name of the key pair, if this instance was launched with an associated key pair.
            returned: always
            type: str
            sample: my-key
        launch_time:
            description: The time the instance was launched.
            returned: always
            type: str
            sample: "2017-03-23T22:51:24+00:00"
        maintenance_options:
            description: Provides information on the recovery and maintenance options of your instance.
            returned: always
            type: dict
            contains:
                auto_recovery:
                    description: Provides information on the current automatic recovery behavior of your instance.
                    type: str
                    sample: default
        metadata_options:
            description: The metadata options for the instance.
            returned: always
            type: complex
            contains:
                http_endpoint:
                    description: Indicates whether the HTTP metadata endpoint on your instances is enabled or disabled.
                    type: str
                    sample: enabled
                http_protocol_ipv6:
                    description: Indicates whether the IPv6 endpoint for the instance metadata service is enabled or disabled.
                    type: str
                    sample: disabled
                http_put_response_hop_limit:
                    description: The maximum number of hops that the metadata token can travel.
                    type: int
                    sample: 1
                http_tokens:
                    description: Indicates whether IMDSv2 is required.
                    type: str
                    sample: optional
                instance_metadata_tags:
                    description: Indicates whether access to instance tags from the instance metadata is enabled or disabled.
                    type: str
                    sample: disabled
                state:
                    description: The state of the metadata option changes.
                    type: str
                    sample: applied
        monitoring:
            description: The monitoring for the instance.
            returned: always
            type: complex
            contains:
                state:
                    description: Indicates whether detailed monitoring is enabled. Otherwise, basic monitoring is enabled.
                    returned: always
                    type: str
                    sample: disabled
        network_interfaces:
            description: One or more network interfaces for the instance.
            returned: always
            type: complex
            contains:
                association:
                    description: The association information for an Elastic IPv4 associated with the network interface.
                    returned: always
                    type: complex
                    contains:
                        ip_owner_id:
                            description: The ID of the owner of the Elastic IP address.
                            returned: always
                            type: str
                            sample: amazon
                        public_dns_name:
                            description: The public DNS name.
                            returned: always
                            type: str
                            sample: ""
                        public_ip:
                            description: The public IP address or Elastic IP address bound to the network interface.
                            returned: always
                            type: str
                            sample: 1.2.3.4
                attachment:
                    description: The network interface attachment.
                    returned: always
                    type: complex
                    contains:
                        attach_time:
                            description: The time stamp when the attachment initiated.
                            returned: always
                            type: str
                            sample: "2017-03-23T22:51:24+00:00"
                        attachment_id:
                            description: The ID of the network interface attachment.
                            returned: always
                            type: str
                            sample: eni-attach-3aff3f
                        delete_on_termination:
                            description: Indicates whether the network interface is deleted when the instance is terminated.
                            returned: always
                            type: bool
                            sample: true
                        device_index:
                            description: The index of the device on the instance for the network interface attachment.
                            returned: always
                            type: int
                            sample: 0
                        network_card_index:
                            description: The index of the network card.
                            returned: always
                            type: int
                            sample: 0
                        status:
                            description: The attachment state.
                            returned: always
                            type: str
                            sample: attached
                description:
                    description: The description.
                    returned: always
                    type: str
                    sample: My interface
                groups:
                    description: One or more security groups.
                    returned: always
                    type: list
                    elements: dict
                    contains:
                        group_id:
                            description: The ID of the security group.
                            returned: always
                            type: str
                            sample: sg-abcdef12
                        group_name:
                            description: The name of the security group.
                            returned: always
                            type: str
                            sample: mygroup
                interface_type:
                    description: The type of network interface.
                    returned: always
                    type: str
                    sample: interface
                ipv6_addresses:
                    description: One or more IPv6 addresses associated with the network interface.
                    returned: always
                    type: list
                    elements: dict
                    contains:
                        ipv6_address:
                            description: The IPv6 address.
                            returned: always
                            type: str
                            sample: "2001:0db8:85a3:0000:0000:8a2e:0370:7334"
                mac_address:
                    description: The MAC address.
                    returned: always
                    type: str
                    sample: "00:11:22:33:44:55"
                network_interface_id:
                    description: The ID of the network interface.
                    returned: always
                    type: str
                    sample: eni-01234567
                owner_id:
                    description: The AWS account ID of the owner of the network interface.
                    returned: always
                    type: str
                    sample: 01234567890
                private_dns_name:
                    description: The private DNS hostname name assigned to the instance.
                    type: str
                    returned: always
                    sample: ip-10-1-0-156.ec2.internal
                private_ip_address:
                    description: The IPv4 address of the network interface within the subnet.
                    returned: always
                    type: str
                    sample: 10.0.0.1
                private_ip_addresses:
                    description: The private IPv4 addresses associated with the network interface.
                    returned: always
                    type: list
                    elements: dict
                    contains:
                        association:
                            description: The association information for an Elastic IP address (IPv4) associated with the network interface.
                            type: complex
                            contains:
                                ip_owner_id:
                                    description: The ID of the owner of the Elastic IP address.
                                    returned: always
                                    type: str
                                    sample: amazon
                                public_dns_name:
                                    description: The public DNS name.
                                    returned: always
                                    type: str
                                    sample: ""
                                public_ip:
                                    description: The public IP address or Elastic IP address bound to the network interface.
                                    returned: always
                                    type: str
                                    sample: 1.2.3.4
                        primary:
                            description: Indicates whether this IPv4 address is the primary private IP address of the network interface.
                            returned: always
                            type: bool
                            sample: true
                        private_dns_name:
                            description: The private DNS hostname name assigned to the instance.
                            type: str
                            returned: always
                            sample: ip-10-1-0-156.ec2.internal
                        private_ip_address:
                            description: The private IPv4 address of the network interface.
                            returned: always
                            type: str
                            sample: 10.0.0.1
                source_dest_check:
                    description: Indicates whether source/destination checking is enabled.
                    returned: always
                    type: bool
                    sample: true
                status:
                    description: The status of the network interface.
                    returned: always
                    type: str
                    sample: in-use
                subnet_id:
                    description: The ID of the subnet for the network interface.
                    returned: always
                    type: str
                    sample: subnet-0123456
                vpc_id:
                    description: The ID of the VPC for the network interface.
                    returned: always
                    type: str
                    sample: vpc-0123456
        placement:
            description: The location where the instance launched, if applicable.
            returned: always
            type: complex
            contains:
                availability_zone:
                    description: The Availability Zone of the instance.
                    returned: always
                    type: str
                    sample: ap-southeast-2a
                group_name:
                    description: The name of the placement group the instance is in (for cluster compute instances).
                    returned: always
                    type: str
                    sample: ""
                tenancy:
                    description: The tenancy of the instance (if the instance is running in a VPC).
                    returned: always
                    type: str
                    sample: default
        platform_details:
            description: The platform details value for the instance.
            returned: always
            type: str
            sample: Linux/UNIX
        private_dns_name:
            description: The private DNS name.
            returned: always
            type: str
            sample: ip-10-0-0-1.ap-southeast-2.compute.internal
        private_dns_name_options:
            description: The options for the instance hostname.
            type: dict
            contains:
                enable_resource_name_dns_a_record:
                    description: Indicates whether to respond to DNS queries for instance hostnames with DNS A records.
                    type: bool
                    sample: false
                enable_resource_name_dns_aaaa_record:
                    description: Indicates whether to respond to DNS queries for instance hostnames with DNS AAAA records.
                    type: bool
                    sample: false
                hostname_type:
                    description: The type of hostname to assign to an instance.
                    type: str
                    sample: ip-name
        private_ip_address:
            description: The IPv4 address of the network interface within the subnet.
            returned: always
            type: str
            sample: 10.0.0.1
        product_codes:
            description: One or more product codes.
            returned: always
            type: list
            elements: dict
            contains:
                product_code_id:
                    description: The product code.
                    returned: always
                    type: str
                    sample: aw0evgkw8ef3n2498gndfgasdfsd5cce
                product_code_type:
                    description: The type of product code.
                    returned: always
                    type: str
                    sample: marketplace
        public_dns_name:
            description: The public DNS name assigned to the instance.
            returned: always
            type: str
            sample:
        public_ip_address:
            description: The public IPv4 address assigned to the instance.
            returned: always
            type: str
            sample: 52.0.0.1
        root_device_name:
            description: The device name of the root device.
            returned: always
            type: str
            sample: /dev/sda1
        root_device_type:
            description: The type of root device used by the AMI.
            returned: always
            type: str
            sample: ebs
        security_groups:
            description: One or more security groups for the instance.
            returned: always
            type: list
            elements: dict
            contains:
                group_id:
                    description: The ID of the security group.
                    returned: always
                    type: str
                    sample: sg-0123456
                group_name:
                    description: The name of the security group.
                    returned: always
                    type: str
                    sample: my-security-group
        source_dest_check:
            description: Indicates whether source/destination checking is enabled.
            returned: always
            type: bool
            sample: true
        state:
            description: The current state of the instance.
            returned: always
            type: complex
            contains:
                code:
                    description: The low byte represents the state.
                    returned: always
                    type: int
                    sample: 16
                name:
                    description: The name of the state.
                    returned: always
                    type: str
                    sample: running
        state_transition_reason:
            description: The reason for the most recent state transition.
            returned: always
            type: str
            sample:
        subnet_id:
            description: The ID of the subnet in which the instance is running.
            returned: always
            type: str
            sample: subnet-00abcdef
        tags:
            description: Any tags assigned to the instance.
            returned: always
            type: dict
            sample:
        virtualization_type:
            description: The type of virtualization of the AMI.
            returned: always
            type: str
            sample: hvm
        vpc_id:
            description: The ID of the VPC the instance is in.
            returned: always
            type: dict
            sample: vpc-0011223344
        attributes:
            description: The details of the instance attribute specified on input.
            returned: when O(include_attributes) is specified
            type: dict
            sample:
                {
                    'disable_api_termination': {
                        'value': True
                    },
                    'ebs_optimized': {
                        'value': True
                    }
                }
            version_added: 6.3.0
"""

import datetime
from typing import Any
from typing import Dict
from typing import List

from ansible.module_utils.common.dict_transformations import camel_dict_to_snake_dict

from ansible_collections.amazon.aws.plugins.module_utils.ec2 import AnsibleEC2Error
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import describe_instance_attribute
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import describe_instances
from ansible_collections.amazon.aws.plugins.module_utils.modules import AnsibleAWSModule
from ansible_collections.amazon.aws.plugins.module_utils.tagging import boto3_tag_list_to_ansible_dict
from ansible_collections.amazon.aws.plugins.module_utils.transformation import ansible_dict_to_boto3_filter_list


def list_ec2_instances(connection, module: AnsibleAWSModule) -> None:
    instance_ids = module.params.get("instance_ids")
    uptime = module.params.get("minimum_uptime")
    filters = ansible_dict_to_boto3_filter_list(module.params.get("filters"))

    reservations = describe_instances(connection, InstanceIds=instance_ids, Filters=filters)
    instances = []
    oldest_launch_time = None

    if uptime:
        timedelta = int(uptime) if uptime else 0
        oldest_launch_time = datetime.datetime.now() - datetime.timedelta(minutes=timedelta)

    # Get instances from reservations
    for reservation in reservations:
        instances.extend(
            [
                instance
                for instance in reservation["Instances"]
                if oldest_launch_time is None or instance["LaunchTime"].replace(tzinfo=None) < oldest_launch_time
            ]
        )

    # include instances attributes
    attributes = module.params.get("include_attributes")
    if attributes:
        for instance in instances:
            instance["attributes"] = describe_instance_attributes(connection, instance["InstanceId"], attributes)

    # Turn the boto3 result in to ansible_friendly_snaked_names
    snaked_instances = [camel_dict_to_snake_dict(instance) for instance in instances]

    # Turn the boto3 result in to ansible friendly tag dictionary
    for instance in snaked_instances:
        instance["tags"] = boto3_tag_list_to_ansible_dict(instance.get("tags", []), "key", "value")

    module.exit_json(instances=snaked_instances)


def describe_instance_attributes(connection, instance_id: str, attributes: List[str]) -> Dict[str, Any]:
    result = {}
    for attr in attributes:
        response = describe_instance_attribute(connection, instance_id=instance_id, attribute=attr)
        for key in response:
            if key not in ("InstanceId", "ResponseMetadata"):
                result[key] = response[key]
    return result


def main():
    instance_attributes = [
        "instanceType",
        "kernel",
        "ramdisk",
        "userData",
        "disableApiTermination",
        "instanceInitiatedShutdownBehavior",
        "rootDeviceName",
        "blockDeviceMapping",
        "productCodes",
        "sourceDestCheck",
        "groupSet",
        "ebsOptimized",
        "sriovNetSupport",
        "enclaveOptions",
        "disableApiStop",
    ]
    argument_spec = dict(
        minimum_uptime=dict(required=False, type="int", default=None, aliases=["uptime"]),
        instance_ids=dict(default=[], type="list", elements="str"),
        filters=dict(default={}, type="dict"),
        include_attributes=dict(type="list", elements="str", aliases=["attributes"], choices=instance_attributes),
    )

    module = AnsibleAWSModule(
        argument_spec=argument_spec,
        mutually_exclusive=[["instance_ids", "filters"]],
        supports_check_mode=True,
    )

    connection = module.client("ec2")
    try:
        list_ec2_instances(connection, module)
    except AnsibleEC2Error as e:
        module.fail_json_aws_error(e)


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2016, Olivier Boukili <boukili.olivier@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: apache2_mod_proxy
author: Olivier Boukili (@oboukili)
short_description: Set and/or get members' attributes of an Apache httpd 2.4 mod_proxy balancer pool
description:
  - Set and/or get members' attributes of an Apache httpd 2.4 mod_proxy balancer pool, using HTTP POST and GET requests. The
    httpd mod_proxy balancer-member status page has to be enabled and accessible, as this module relies on parsing this page.
extends_documentation_fragment:
  - community.general._attributes
requirements:
  - Python package C(beautifulsoup4)
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  balancer_url_suffix:
    type: str
    description:
      - Suffix of the balancer pool URL required to access the balancer pool status page (for example V(balancer_vhost[:port]/balancer_url_suffix)).
    default: /balancer-manager/
  balancer_vhost:
    type: str
    description:
      - (IPv4|IPv6|FQDN):port of the Apache httpd 2.4 mod_proxy balancer pool.
    required: true
  member_host:
    type: str
    description:
      - (IPv4|IPv6|FQDN) of the balancer member to get or to set attributes to. Port number is autodetected and should not
        be specified here.
      - If undefined, the M(community.general.apache2_mod_proxy) module returns a members list of dictionaries of all the
        current balancer pool members' attributes.
  state:
    type: list
    elements: str
    choices: [present, absent, enabled, disabled, drained, hot_standby, ignore_errors]
    description:
      - Desired state of the member host.
      - States can be simultaneously invoked by separating them with a comma (for example V(state=drained,ignore_errors)),
        but it is recommended to specify them as a proper YAML list.
      - States V(present) and V(absent) must be used without any other state.
  tls:
    description:
      - Use https to access balancer management page.
    type: bool
    default: false
  validate_certs:
    description:
      - Validate ssl/tls certificates.
    type: bool
    default: true
"""

EXAMPLES = r"""
- name: Get all current balancer pool members attributes
  community.general.apache2_mod_proxy:
    balancer_vhost: 10.0.0.2

- name: Get a specific member attributes
  community.general.apache2_mod_proxy:
    balancer_vhost: myws.mydomain.org
    balancer_suffix: /lb/
    member_host: node1.myws.mydomain.org

# Enable all balancer pool members:
- name: Get attributes
  community.general.apache2_mod_proxy:
    balancer_vhost: '{{ myloadbalancer_host }}'
  register: result

- name: Enable all balancer pool members
  community.general.apache2_mod_proxy:
    balancer_vhost: '{{ myloadbalancer_host }}'
    member_host: '{{ item.host }}'
    state: present
  with_items: '{{ result.members }}'

# Gracefully disable a member from a loadbalancer node:
- name: Step 1
  community.general.apache2_mod_proxy:
    balancer_vhost: '{{ vhost_host }}'
    member_host: '{{ member.host }}'
    state: drained
  delegate_to: myloadbalancernode

- name: Step 2
  ansible.builtin.wait_for:
    host: '{{ member.host }}'
    port: '{{ member.port }}'
    state: drained
  delegate_to: myloadbalancernode

- name: Step 3
  community.general.apache2_mod_proxy:
    balancer_vhost: '{{ vhost_host }}'
    member_host: '{{ member.host }}'
    state: absent
  delegate_to: myloadbalancernode
"""

RETURN = r"""
member:
  description: Specific balancer member information dictionary, returned when the module is invoked with O(member_host) parameter.
  type: dict
  returned: success
  sample:
    {
      "attributes": {
        "Busy": "0",
        "Elected": "42",
        "Factor": "1",
        "From": "136K",
        "Load": "0",
        "Route": null,
        "RouteRedir": null,
        "Set": "0",
        "Status": "Init Ok ",
        "To": " 47K",
        "Worker URL": null
      },
      "balancer_url": "http://10.10.0.2/balancer-manager/",
      "host": "10.10.0.20",
      "management_url": "http://10.10.0.2/lb/?b=mywsbalancer&w=http://10.10.0.20:8080/ws&nonce=8925436c-79c6-4841-8936-e7d13b79239b",
      "path": "/ws",
      "port": 8080,
      "protocol": "http",
      "status": {
        "disabled": false,
        "drained": false,
        "hot_standby": false,
        "ignore_errors": false
      }
    }
members:
  description: List of member (defined above) dictionaries, returned when the module is invoked with no O(member_host) and
    O(state) args.
  returned: success
  type: list
  sample:
    [
      {
        "attributes": {
          "Busy": "0",
          "Elected": "42",
          "Factor": "1",
          "From": "136K",
          "Load": "0",
          "Route": null,
          "RouteRedir": null,
          "Set": "0",
          "Status": "Init Ok ",
          "To": " 47K",
          "Worker URL": null
        },
        "balancer_url": "http://10.10.0.2/balancer-manager/",
        "host": "10.10.0.20",
        "management_url": "http://10.10.0.2/lb/?b=mywsbalancer&w=http://10.10.0.20:8080/ws&nonce=8925436c-79c6-4841-8936-e7d13b79239b",
        "path": "/ws",
        "port": 8080,
        "protocol": "http",
        "status": {
          "disabled": false,
          "drained": false,
          "hot_standby": false,
          "ignore_errors": false
        }
      },
      {
        "attributes": {
          "Busy": "0",
          "Elected": "42",
          "Factor": "1",
          "From": "136K",
          "Load": "0",
          "Route": null,
          "RouteRedir": null,
          "Set": "0",
          "Status": "Init Ok ",
          "To": " 47K",
          "Worker URL": null
        },
        "balancer_url": "http://10.10.0.2/balancer-manager/",
        "host": "10.10.0.21",
        "management_url": "http://10.10.0.2/lb/?b=mywsbalancer&w=http://10.10.0.21:8080/ws&nonce=8925436c-79c6-4841-8936-e7d13b79239b",
        "path": "/ws",
        "port": 8080,
        "protocol": "http",
        "status": {
          "disabled": false,
          "drained": false,
          "hot_standby": false,
          "ignore_errors": false
        }
      }
    ]
"""

import re

from ansible.module_utils.common.text.converters import to_text
from ansible.module_utils.urls import fetch_url

from ansible_collections.community.general.plugins.module_utils import _deps as deps
from ansible_collections.community.general.plugins.module_utils._module_helper import (
    ModuleHelper,
    ModuleHelperException,
)

with deps.declare("beautifulsoup4"):
    from bs4 import BeautifulSoup

# balancer member attributes extraction regexp:
EXPRESSION = re.compile(
    to_text(r"(b=([\w\.\-]+)&w=(https?|ajp|wss?|ftp|[sf]cgi)://([\w\.\-]+):?(\d*)([/\w\.\-]*)&?[\w\-\=]*)")
)
# Apache2 server version extraction regexp:
APACHE_VERSION_EXPRESSION = re.compile(to_text(r"SERVER VERSION: APACHE/([\d.]+)"))


def find_all(where, what):
    return where.find_all(what)


def regexp_extraction(string, _regexp, groups=1):
    """Returns the capture group (default=1) specified in the regexp, applied to the string"""
    regexp_search = _regexp.search(string)
    if regexp_search:
        if regexp_search.group(groups) != "":
            return regexp_search.group(groups)
    return None


class BalancerMember:
    """Apache 2.4 mod_proxy LB balancer member.
    attributes:
        read-only:
            host -> member host (string),
            management_url -> member management url (string),
            protocol -> member protocol (string)
            port -> member port (string),
            path -> member location (string),
            balancer_url -> url of this member's parent balancer (string),
            attributes -> whole member attributes (dictionary)
            module -> ansible module instance (AnsibleModule object).
        writable:
            status -> status of the member (dictionary)
    """

    def __init__(self, management_url, balancer_url, module):
        self.host = regexp_extraction(management_url, EXPRESSION, 4)
        self.management_url = management_url
        self.protocol = regexp_extraction(management_url, EXPRESSION, 3)
        self.port = regexp_extraction(management_url, EXPRESSION, 5)
        self.path = regexp_extraction(management_url, EXPRESSION, 6)
        self.balancer_url = balancer_url
        self.module = module

    def get_member_attributes(self):
        """Returns a dictionary of a balancer member's attributes."""

        resp, info = fetch_url(self.module, self.management_url, headers={"Referer": self.management_url})

        if info["status"] != 200:
            raise ModuleHelperException(f"Could not get balancer_member_page, check for connectivity! {info}")

        try:
            soup = BeautifulSoup(resp)
        except TypeError as exc:
            raise ModuleHelperException(f"Cannot parse balancer_member_page HTML! {exc}") from exc

        subsoup = find_all(find_all(soup, "table")[1], "tr")
        keys = find_all(subsoup[0], "th")
        for valuesset in subsoup[1::1]:
            if re.search(pattern=self.host, string=str(valuesset)):
                values = find_all(valuesset, "td")
                return {keys[x].string: values[x].string for x in range(0, len(keys))}

    def get_member_status(self):
        """Returns a dictionary of a balancer member's status attributes."""
        status_mapping = {"disabled": "Dis", "drained": "Drn", "hot_standby": "Stby", "ignore_errors": "Ign"}
        actual_status = self.attributes["Status"]
        status = {mode: patt in actual_status for mode, patt in status_mapping.items()}
        return status

    def set_member_status(self, values):
        """Sets a balancer member's status attributes amongst pre-mapped values."""
        values_mapping = {
            "disabled": "&w_status_D",
            "drained": "&w_status_N",
            "hot_standby": "&w_status_H",
            "ignore_errors": "&w_status_I",
        }

        request_body = regexp_extraction(self.management_url, EXPRESSION, 1)
        values_url = "".join(f"{url_param}={1 if values[mode] else 0}" for mode, url_param in values_mapping.items())
        request_body = f"{request_body}{values_url}"

        response, info = fetch_url(
            self.module, self.management_url, data=request_body, headers={"Referer": self.management_url}
        )
        if info["status"] != 200:
            raise ModuleHelperException(f"Could not set the member status! {self.host} {info['status']}")

    attributes = property(get_member_attributes)
    status = property(get_member_status, set_member_status)

    def as_dict(self):
        return {
            "host": self.host,
            "status": self.status,
            "protocol": self.protocol,
            "port": self.port,
            "path": self.path,
            "attributes": self.attributes,
            "management_url": self.management_url,
            "balancer_url": self.balancer_url,
        }


class Balancer:
    """Apache httpd 2.4 mod_proxy balancer object"""

    def __init__(self, module, host, suffix, tls=False):
        proto = "https" if tls else "http"
        self.base_url = f"{proto}://{host}"
        self.url = f"{proto}://{host}{suffix}"
        self.module = module
        self.page = self.fetch_balancer_page()

    def fetch_balancer_page(self):
        """Returns the balancer management html page as a string for later parsing."""
        resp, info = fetch_url(self.module, self.url)
        if info["status"] != 200:
            raise ModuleHelperException(f"Could not get balancer page! HTTP status response: {info['status']}")

        content = to_text(resp.read())
        apache_version = regexp_extraction(content.upper(), APACHE_VERSION_EXPRESSION, 1)
        if not apache_version:
            raise ModuleHelperException("Could not get the Apache server version from the balancer-manager")

        if not re.search(pattern=r"2\.4\.[\d]*", string=apache_version):
            raise ModuleHelperException(
                f"This module only acts on an Apache2 2.4+ instance, current Apache2 version: {apache_version}"
            )
        return content

    def get_balancer_members(self):
        """Returns members of the balancer as a generator object for later iteration."""
        try:
            soup = BeautifulSoup(self.page)
        except TypeError as e:
            raise ModuleHelperException(f"Cannot parse balancer page HTML! {self.page}") from e

        elements = find_all(soup, "a")
        for element in elements[1::1]:
            balancer_member_suffix = element.get("href")
            if not balancer_member_suffix:
                raise ModuleHelperException("Argument 'balancer_member_suffix' is empty!")

            yield BalancerMember(self.base_url + balancer_member_suffix, self.url, self.module)

    members = property(get_balancer_members)


class ApacheModProxy(ModuleHelper):
    """Initiates module."""

    module = dict(
        argument_spec=dict(
            balancer_vhost=dict(required=True, type="str"),
            balancer_url_suffix=dict(default="/balancer-manager/", type="str"),
            member_host=dict(type="str"),
            state=dict(
                type="list",
                elements="str",
                choices=["present", "absent", "enabled", "disabled", "drained", "hot_standby", "ignore_errors"],
            ),
            tls=dict(default=False, type="bool"),
            validate_certs=dict(default=True, type="bool"),
        ),
        supports_check_mode=True,
    )

    def __init_module__(self):
        deps.validate(self.module)

        if len(self.vars.state or []) > 1 and ("present" in self.vars.state or "enabled" in self.vars.state):
            self.do_raise(msg="states present/enabled are mutually exclusive with other states!")

        self.mybalancer = Balancer(
            self.module, self.vars.balancer_vhost, self.vars.balancer_url_suffix, tls=self.vars.tls
        )

    def __run__(self):
        if self.vars.member_host is None:
            self.vars.members = [member.as_dict() for member in self.mybalancer.members]
        else:
            member_exists = False
            member_status = {"disabled": False, "drained": False, "hot_standby": False, "ignore_errors": False}
            for mode in member_status:
                for state in self.vars.state or []:
                    if mode == state:
                        member_status[mode] = True
                    elif mode == "disabled" and state == "absent":
                        member_status[mode] = True

            for member in self.mybalancer.members:
                if str(member.host) == self.vars.member_host:
                    member_exists = True
                    if self.vars.state is not None:
                        member_status_before = member.status
                        if not self.check_mode:
                            member_status_after = member.status = member_status
                        else:
                            member_status_after = member_status
                        self.changed |= member_status_before != member_status_after
                    self.vars.member = member.as_dict()

            if not member_exists:
                self.do_raise(
                    msg=f"{self.vars.member_host} is not a member of the balancer {self.vars.balancer_vhost}!"
                )


def main():
    ApacheModProxy.execute()


if __name__ == "__main__":
    main()



# Copyright (c) 2012, Jan-Piet Mens <jpmens(at)gmail.com>
# Copyright (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
from __future__ import annotations

DOCUMENTATION = r"""
name: redis
author:
  - Jan-Piet Mens (@jpmens) <jpmens(at)gmail.com>
  - Ansible Core Team
short_description: Fetch data from Redis
description:
  - This lookup returns a list of results from a Redis DB corresponding to a list of items given to it.
requirements:
  - redis (python library https://github.com/andymccurdy/redis-py/)
options:
  _terms:
    description: List of keys to query.
    type: list
    elements: string
  host:
    description: Location of Redis host.
    type: string
    default: '127.0.0.1'
    env:
      - name: ANSIBLE_REDIS_HOST
    ini:
      - section: lookup_redis
        key: host
  port:
    description: Port on which Redis is listening on.
    default: 6379
    type: int
    env:
      - name: ANSIBLE_REDIS_PORT
    ini:
      - section: lookup_redis
        key: port
  socket:
    description: Path to socket on which to query Redis, this option overrides host and port options when set.
    type: path
    env:
      - name: ANSIBLE_REDIS_SOCKET
    ini:
      - section: lookup_redis
        key: socket
"""

EXAMPLES = r"""
- name: query redis for somekey (default or configured settings used)
  ansible.builtin.debug:
    msg: "{{ lookup('community.general.redis', 'somekey') }}"

- name: query redis for list of keys and non-default host and port
  ansible.builtin.debug:
    msg: "{{ lookup('community.general.redis', item, host='myredis.internal.com', port=2121) }}"
  loop: '{{list_of_redis_keys}}'

- name: use list directly
  ansible.builtin.debug:
    msg: "{{ lookup('community.general.redis', 'key1', 'key2', 'key3') }}"

- name: use list directly with a socket
  ansible.builtin.debug:
    msg: "{{ lookup('community.general.redis', 'key1', 'key2', socket='/var/tmp/redis.sock') }}"
"""

RETURN = r"""
_raw:
  description: Value(s) stored in Redis.
  type: list
  elements: str
"""

HAVE_REDIS = False
try:
    import redis

    HAVE_REDIS = True
except ImportError:
    pass

from ansible.errors import AnsibleError
from ansible.module_utils.common.text.converters import to_text
from ansible.plugins.lookup import LookupBase

from ansible_collections.community.general.plugins.plugin_utils._lookup import check_for_wrong_terms


class LookupModule(LookupBase):
    def run(self, terms, variables, **kwargs):
        if not HAVE_REDIS:
            raise AnsibleError("Can't LOOKUP(redis_kv): module redis is not installed")

        # get options
        self.set_options(direct=kwargs)
        check_for_wrong_terms(self, direct=kwargs)

        # setup connection
        host = self.get_option("host")
        port = self.get_option("port")
        socket = self.get_option("socket")
        if socket is None:
            conn = redis.Redis(host=host, port=port)
        else:
            conn = redis.Redis(unix_socket_path=socket)

        ret = []
        for term in terms:
            try:
                res = conn.get(term)
                if res is None:
                    res = ""
                ret.append(to_text(res))
            except Exception as e:
                # connection failed or key not found
                raise AnsibleError(f"Encountered exception while fetching {term}: {e}") from e
        return ret



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: mq_user
version_added: 6.0.0
short_description: Manage users in existing Amazon MQ broker
description:
  - Manage Amazon MQ users.
  - Pending changes are taking into account for idempotency.
author:
  - FCO (@fotto)
options:
  broker_id:
    description:
      - The ID of the MQ broker to work on.
    type: str
    required: true
  username:
    description:
      - The name of the user to create/update/delete.
    type: str
    required: true
  state:
    description:
      - Create/Update vs Delete of user.
    default: present
    choices: [ 'present', 'absent' ]
    type: str
  console_access:
    description:
      - Whether the user can access the MQ Console.
      - Defaults to C(false) on creation.
    type: bool
  groups:
    description:
      - Set group memberships for user.
      - Defaults to C([]) on creation.
    type: list
    elements: str
  password:
    description:
      - Set password for user.
      - Defaults to a random password on creation.
      - Ignored unless I(allow_pw_update=true).
    type: str
  allow_pw_update:
    description:
      - When I(allow_pw_update=true) and I(password) is set, the password
        will always be updated for the user.
    default: false
    type: bool
extends_documentation_fragment:
  - amazon.aws.boto3
  - amazon.aws.common.modules
  - amazon.aws.region.modules
"""

EXAMPLES = r"""
- name: create/update user - set provided password if user doesn't exist, yet
  community.aws.mq_user:
    state: present
    broker_id: "aws-mq-broker-id"
    username: "sample_user1"
    console_access: false
    groups: ["g1", "g2"]
    password: "plain-text-password"

- name: allow console access and update group list - relying on default state
  community.aws.mq_user:
    broker_id: "aws-mq-broker-id"
    username: "sample_user1"
    region: "{{ aws_region }}"
    console_access: true
    groups: ["g1", "g2", "g3"]

- name: remove user - setting all credentials explicitly
  community.aws.mq_user:
    state: absent
    broker_id: "aws-mq-broker-id"
    username: "other_user"
"""

RETURN = r"""
user:
    description:
      - just echos the username
      - "only present when state=present"
    type: str
    returned: success
"""

import secrets

try:
    import botocore
except ImportError:
    # handled by AnsibleAWSModule
    pass

from ansible.module_utils.common.dict_transformations import camel_dict_to_snake_dict

from ansible_collections.amazon.aws.plugins.module_utils.botocore import is_boto3_error_code
from ansible_collections.amazon.aws.plugins.module_utils.modules import AnsibleAWSModule

CREATE_DEFAULTS = {
    "console_access": False,
    "groups": [],
}


def _group_change_required(user_response, requested_groups):
    current_groups = []
    if "Groups" in user_response:
        current_groups = user_response["Groups"]
    elif "Pending" in user_response:
        # to support automatic testing without broker reboot
        current_groups = user_response["Pending"]["Groups"]
    if len(current_groups) != len(requested_groups):
        return True
    if len(current_groups) != len(set(current_groups) & set(requested_groups)):
        return True
    #
    return False


def _console_access_change_required(user_response, requested_boolean):
    current_boolean = CREATE_DEFAULTS["console_access"]
    if "ConsoleAccess" in user_response:
        current_boolean = user_response["ConsoleAccess"]
    elif "Pending" in user_response:
        # to support automatic testing without broker reboot
        current_boolean = user_response["Pending"]["ConsoleAccess"]
    #
    return current_boolean != requested_boolean


def generate_password():
    return secrets.token_hex(20)


# returns API response object
def _create_user(conn, module):
    kwargs = {"BrokerId": module.params["broker_id"], "Username": module.params["username"]}
    if "groups" in module.params and module.params["groups"] is not None:
        kwargs["Groups"] = module.params["groups"]
    else:
        kwargs["Groups"] = CREATE_DEFAULTS["groups"]
    if "password" in module.params and module.params["password"]:
        kwargs["Password"] = module.params["password"]
    else:
        kwargs["Password"] = generate_password()
    if "console_access" in module.params and module.params["console_access"] is not None:
        kwargs["ConsoleAccess"] = module.params["console_access"]
    else:
        kwargs["ConsoleAccess"] = CREATE_DEFAULTS["console_access"]
    try:
        response = conn.create_user(**kwargs)
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
        module.fail_json_aws(e, msg="Couldn't create user")
    return response


# returns API response object
def _update_user(conn, module, kwargs):
    try:
        response = conn.update_user(**kwargs)
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
        module.fail_json_aws(e, msg="Couldn't update user")
    return response


def get_matching_user(conn, module, broker_id, username):
    try:
        response = conn.describe_user(BrokerId=broker_id, Username=username)
    except is_boto3_error_code("NotFoundException"):
        return None
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
        module.fail_json_aws(e, msg="Couldn't get user details")
    return response


def ensure_user_present(conn, module):
    user = get_matching_user(conn, module, module.params["broker_id"], module.params["username"])
    changed = False

    if user is None:
        if not module.check_mode:
            _response = _create_user(conn, module)
        changed = True
    else:
        kwargs = {}
        if "groups" in module.params and module.params["groups"] is not None:
            if _group_change_required(user, module.params["groups"]):
                kwargs["Groups"] = module.params["groups"]
        if "console_access" in module.params and module.params["console_access"] is not None:
            if _console_access_change_required(user, module.params["console_access"]):
                kwargs["ConsoleAccess"] = module.params["console_access"]
        if "password" in module.params and module.params["password"]:
            if "allow_pw_update" in module.params and module.params["allow_pw_update"]:
                kwargs["Password"] = module.params["password"]
        if len(kwargs) == 0:
            changed = False
        else:
            if not module.check_mode:
                kwargs["BrokerId"] = module.params["broker_id"]
                kwargs["Username"] = module.params["username"]
                _update_user(conn, module, kwargs)
            #
            changed = True
    #
    user = get_matching_user(conn, module, module.params["broker_id"], module.params["username"])

    return {"changed": changed, "user": camel_dict_to_snake_dict(user, ignore_list=["Tags"])}


def ensure_user_absent(conn, module):
    user = get_matching_user(conn, module, module.params["broker_id"], module.params["username"])
    result = {"changed": False}
    if user is None:
        return result
    # better support for testing
    if "Pending" in user and "PendingChange" in user["Pending"] and user["Pending"]["PendingChange"] == "DELETE":
        return result

    result = {"changed": True}
    if module.check_mode:
        return result

    try:
        conn.delete_user(BrokerId=user["BrokerId"], Username=user["Username"])
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
        module.fail_json_aws(e, msg="Couldn't delete user")

    return result


def main():
    argument_spec = dict(
        broker_id=dict(required=True, type="str"),
        username=dict(required=True, type="str"),
        console_access=dict(required=False, type="bool"),
        groups=dict(required=False, type="list", elements="str"),
        password=dict(required=False, type="str", no_log=True),
        allow_pw_update=dict(default=False, required=False, type="bool"),
        state=dict(default="present", choices=["present", "absent"]),
    )

    module = AnsibleAWSModule(argument_spec=argument_spec, supports_check_mode=True)

    connection = module.client("mq")

    state = module.params.get("state")

    try:
        if state == "present":
            result = ensure_user_present(connection, module)
        elif state == "absent":
            result = ensure_user_absent(connection, module)
    except botocore.exceptions.ClientError as e:
        module.fail_json_aws(e)

    module.exit_json(**result)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2018, Loic BLOT (@nerzhul) <loic.blot@unix-experience.fr>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

# This module is sponsored by E.T.A.I. (www.etai.fr)

DOCUMENTATION = r"""
---
module: storagegateway_info
version_added: 1.0.0
short_description: Fetch AWS Storage Gateway information
description:
  - Fetch AWS Storage Gateway information
  - Prior to release 5.0.0 this module was called C(community.aws.aws_sgw_info).
    The usage did not change.
author:
  - Loic Blot (@nerzhul) <loic.blot@unix-experience.fr>
options:
  gather_local_disks:
    description:
      - Gather local disks attached to the storage gateway.
    type: bool
    required: false
    default: true
  gather_tapes:
    description:
      - Gather tape information for storage gateways in tape mode.
    type: bool
    required: false
    default: true
  gather_file_shares:
    description:
      - Gather file share information for storage gateways in s3 mode.
    type: bool
    required: false
    default: true
  gather_volumes:
    description:
      - Gather volume information for storage gateways in iSCSI (cached & stored) modes.
    type: bool
    required: false
    default: true
extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.boto3
"""

RETURN = r"""
gateways:
  description: list of gateway objects
  returned: always
  type: complex
  contains:
    gateway_arn:
      description: "Storage Gateway ARN"
      returned: always
      type: str
      sample: "arn:aws:storagegateway:eu-west-1:123456789012:gateway/sgw-9999F888"
    gateway_id:
      description: "Storage Gateway ID"
      returned: always
      type: str
      sample: "sgw-9999F888"
    gateway_name:
      description: "Storage Gateway friendly name"
      returned: always
      type: str
      sample: "my-sgw-01"
    gateway_operational_state:
      description: "Storage Gateway operational state"
      returned: always
      type: str
      sample: "ACTIVE"
    gateway_type:
      description: "Storage Gateway type"
      returned: always
      type: str
      sample: "FILE_S3"
    file_shares:
      description: "Storage gateway file shares"
      returned: when gateway_type == "FILE_S3"
      type: complex
      contains:
        file_share_arn:
          description: "File share ARN"
          returned: always
          type: str
          sample: "arn:aws:storagegateway:eu-west-1:123456789012:share/share-AF999C88"
        file_share_id:
          description: "File share ID"
          returned: always
          type: str
          sample: "share-AF999C88"
        file_share_status:
          description: "File share status"
          returned: always
          type: str
          sample: "AVAILABLE"
    tapes:
        description: "Storage Gateway tapes"
        returned: when gateway_type == "VTL"
        type: complex
        contains:
          tape_arn:
            description: "Tape ARN"
            returned: always
            type: str
            sample: "arn:aws:storagegateway:eu-west-1:123456789012:tape/tape-AF999C88"
          tape_barcode:
            description: "Tape ARN"
            returned: always
            type: str
            sample: "tape-AF999C88"
          tape_size_in_bytes:
            description: "Tape ARN"
            returned: always
            type: int
            sample: 555887569
          tape_status:
            description: "Tape ARN"
            returned: always
            type: str
            sample: "AVAILABLE"
    local_disks:
      description: "Storage gateway local disks"
      returned: always
      type: complex
      contains:
        disk_allocation_type:
          description: "Disk allocation type"
          returned: always
          type: str
          sample: "CACHE STORAGE"
        disk_id:
          description: "Disk ID on the system"
          returned: always
          type: str
          sample: "pci-0000:00:1f.0"
        disk_node:
          description: "Disk parent block device"
          returned: always
          type: str
          sample: "/dev/sdb"
        disk_path:
          description: "Disk path used for the cache"
          returned: always
          type: str
          sample: "/dev/nvme1n1"
        disk_size_in_bytes:
          description: "Disk size in bytes"
          returned: always
          type: int
          sample: 107374182400
        disk_status:
          description: "Disk status"
          returned: always
          type: str
          sample: "present"
"""

EXAMPLES = r"""
# Note: These examples do not set authentication details, see the AWS Guide for details.

- name: "Get AWS storage gateway information"
  community.aws.storagegateway_info:

- name: "Get AWS storage gateway information for region eu-west-3"
  community.aws.storagegateway_info:
    region: eu-west-3
"""

try:
    from botocore.exceptions import BotoCoreError
    from botocore.exceptions import ClientError
except ImportError:
    pass  # caught by AnsibleAWSModule

from ansible.module_utils.common.dict_transformations import camel_dict_to_snake_dict

from ansible_collections.community.aws.plugins.module_utils.modules import AnsibleCommunityAWSModule as AnsibleAWSModule


class SGWInformationManager(object):
    def __init__(self, client, module):
        self.client = client
        self.module = module
        self.name = self.module.params.get("name")

    def fetch(self):
        gateways = self.list_gateways()
        for gateway in gateways:
            if self.module.params.get("gather_local_disks"):
                self.list_local_disks(gateway)
            # File share gateway
            if gateway["gateway_type"] == "FILE_S3" and self.module.params.get("gather_file_shares"):
                self.list_gateway_file_shares(gateway)
            # Volume tape gateway
            elif gateway["gateway_type"] == "VTL" and self.module.params.get("gather_tapes"):
                self.list_gateway_vtl(gateway)
            # iSCSI gateway
            elif gateway["gateway_type"] in ["CACHED", "STORED"] and self.module.params.get("gather_volumes"):
                self.list_gateway_volumes(gateway)

        self.module.exit_json(gateways=gateways)

    def list_gateways(self):
        """
        List all storage gateways for the AWS endpoint.
        """
        try:
            paginator = self.client.get_paginator("list_gateways")
            response = paginator.paginate(
                PaginationConfig={
                    "PageSize": 100,
                }
            ).build_full_result()

            gateways = []
            for gw in response["Gateways"]:
                gateways.append(camel_dict_to_snake_dict(gw))

            return gateways

        except (BotoCoreError, ClientError) as e:
            self.module.fail_json_aws(e, msg="Couldn't list storage gateways")

    @staticmethod
    def _read_gateway_fileshare_response(fileshares, aws_reponse):
        """
        Read file share objects from AWS API response.
        Drop the gateway_arn attribute from response, as it will be duplicate with parent object.
        """
        for share in aws_reponse["FileShareInfoList"]:
            share_obj = camel_dict_to_snake_dict(share)
            if "gateway_arn" in share_obj:
                del share_obj["gateway_arn"]
            fileshares.append(share_obj)

        return aws_reponse["NextMarker"] if "NextMarker" in aws_reponse else None

    def list_gateway_file_shares(self, gateway):
        """
        List file shares attached to AWS storage gateway when in S3 mode.
        """
        try:
            response = self.client.list_file_shares(GatewayARN=gateway["gateway_arn"], Limit=100)

            gateway["file_shares"] = []
            marker = self._read_gateway_fileshare_response(gateway["file_shares"], response)

            while marker is not None:
                response = self.client.list_file_shares(GatewayARN=gateway["gateway_arn"], Marker=marker, Limit=100)

                marker = self._read_gateway_fileshare_response(gateway["file_shares"], response)
        except (BotoCoreError, ClientError) as e:
            self.module.fail_json_aws(e, msg="Couldn't list gateway file shares")

    def list_local_disks(self, gateway):
        """
        List storage gateway local disks
        """
        try:
            gateway["local_disks"] = [
                camel_dict_to_snake_dict(disk)
                for disk in self.client.list_local_disks(GatewayARN=gateway["gateway_arn"])["Disks"]
            ]
        except (BotoCoreError, ClientError) as e:
            self.module.fail_json_aws(e, msg="Couldn't list storage gateway local disks")

    @staticmethod
    def _read_gateway_tape_response(tapes, aws_response):
        """
        Read tape objects from AWS API response.
        Drop the gateway_arn attribute from response, as it will be duplicate with parent object.
        """
        for tape in aws_response["TapeInfos"]:
            tape_obj = camel_dict_to_snake_dict(tape)
            if "gateway_arn" in tape_obj:
                del tape_obj["gateway_arn"]
            tapes.append(tape_obj)

        return aws_response["Marker"] if "Marker" in aws_response else None

    def list_gateway_vtl(self, gateway):
        """
        List VTL & VTS attached to AWS storage gateway in VTL mode
        """
        try:
            response = self.client.list_tapes(Limit=100)

            gateway["tapes"] = []
            marker = self._read_gateway_tape_response(gateway["tapes"], response)

            while marker is not None:
                response = self.client.list_tapes(Marker=marker, Limit=100)

                marker = self._read_gateway_tape_response(gateway["tapes"], response)
        except (BotoCoreError, ClientError) as e:
            self.module.fail_json_aws(e, msg="Couldn't list storage gateway tapes")

    def list_gateway_volumes(self, gateway):
        """
        List volumes attached to AWS storage gateway in CACHED or STORAGE mode
        """
        try:
            paginator = self.client.get_paginator("list_volumes")
            response = paginator.paginate(
                GatewayARN=gateway["gateway_arn"],
                PaginationConfig={
                    "PageSize": 100,
                },
            ).build_full_result()

            gateway["volumes"] = []
            for volume in response["VolumeInfos"]:
                volume_obj = camel_dict_to_snake_dict(volume)
                if "gateway_arn" in volume_obj:
                    del volume_obj["gateway_arn"]
                if "gateway_id" in volume_obj:
                    del volume_obj["gateway_id"]

                gateway["volumes"].append(volume_obj)
        except (BotoCoreError, ClientError) as e:
            self.module.fail_json_aws(e, msg="Couldn't list storage gateway volumes")


def main():
    argument_spec = dict(
        gather_local_disks=dict(type="bool", default=True),
        gather_tapes=dict(type="bool", default=True),
        gather_file_shares=dict(type="bool", default=True),
        gather_volumes=dict(type="bool", default=True),
    )

    module = AnsibleAWSModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
    )

    client = module.client("storagegateway")

    if client is None:  # this should never happen
        module.fail_json(msg="Unknown error, failed to create storagegateway client, no information available.")

    SGWInformationManager(client, module).fetch()


if __name__ == "__main__":
    main()



#
# -*- coding: utf-8 -*-
# Copyright 2021 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
#

"""
The replace_keys plugin code
"""
from __future__ import absolute_import, division, print_function


__metaclass__ = type

import re

from ansible.errors import AnsibleFilterError


def _raise_error(msg):
    """Raise an error message, prepend with filter name
    :param msg: The message
    :type msg: str
    :raises: AnsibleError
    """
    error = "Error when using plugin 'replace_keys': {msg}".format(msg=msg)
    raise AnsibleFilterError(error)


def replace_keys_from_dict_n_list(data, target, matching_parameter):
    if isinstance(data, dict):
        for key in target:
            for k in list(data.keys()):
                if matching_parameter == "regex":
                    if re.match(key.get("before"), k):
                        data[key.get("after")] = data.pop(k)
                elif matching_parameter == "starts_with":
                    if k.startswith(key.get("before")):
                        data[key.get("after")] = data.pop(k)
                elif matching_parameter == "ends_with":
                    if k.endswith(key.get("before")):
                        data[key.get("after")] = data.pop(k)
                else:
                    if k == key.get("before"):
                        data[key.get("after")] = data.pop(k)
        for k, v in data.items():
            replace_keys_from_dict_n_list(v, target, matching_parameter)
    elif isinstance(data, list):
        for i in data:
            replace_keys_from_dict_n_list(i, target, matching_parameter)
    return data


def replace_keys(data, target, matching_parameter="equality"):
    """replaces specific keys with mentioned after data"
    :param data: The data passed in (data|replace_keys(...))
    :type data: raw
    :param target: List of keys on with operation is to be performed
    :type data: list
    :type elements: string
    :param matching_parameter: matching type of the target keys with data keys
    :type data: list
    :type elements: dict
    """
    if not isinstance(data, (list, dict)):
        _raise_error("Input is not valid for replace operation")
    data = replace_keys_from_dict_n_list(data, target, matching_parameter)
    return data



#!/usr/bin/python

# Copyright 2013 Matt Coddington <coddington@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: newrelic_deployment
author: "Matt Coddington (@mcodd)"
short_description: Notify New Relic about app deployments
description:
  - Notify New Relic about app deployments (see U(https://docs.newrelic.com/docs/apm/new-relic-apm/maintenance/record-monitor-deployments/)).
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  token:
    type: str
    description:
      - API token to place in the Api-Key header.
    required: true
  app_name:
    type: str
    description:
      - The value of C(app_name) in the C(newrelic.yml) file used by the application.
      - One of O(app_name) or O(application_id) is required.
  application_id:
    type: str
    description:
      - The application ID found in the metadata of the application in APM.
      - One of O(app_name) or O(application_id) is required.
  changelog:
    type: str
    description:
      - A list of changes for this deployment.
  description:
    type: str
    description:
      - Text annotation for the deployment - notes for you.
  revision:
    type: str
    description:
      - A revision number (for example, git commit SHA).
    required: true
  user:
    type: str
    description:
      - The name of the user/process that triggered this deployment.
  validate_certs:
    description:
      - If V(false), SSL certificates are not validated. This should only be used on personally controlled sites using self-signed
        certificates.
    default: true
    type: bool
  app_name_exact_match:
    type: bool
    description:
      - If this flag is set to V(true) then the application ID lookup by name would only work for an exact match. If set to
        V(false) it returns the first result.
    default: false
    version_added: 7.5.0
requirements: []
"""

EXAMPLES = r"""
- name: Notify New Relic about an app deployment
  community.general.newrelic_deployment:
    token: AAAAAA
    app_name: myapp
    user: ansible deployment
    revision: '1.0'
"""

import json
from urllib.parse import quote

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.urls import fetch_url

# ===========================================
# Module execution.
#


def main():
    module = AnsibleModule(
        argument_spec=dict(
            token=dict(required=True, no_log=True),
            app_name=dict(),
            application_id=dict(),
            changelog=dict(),
            description=dict(),
            revision=dict(required=True),
            user=dict(),
            validate_certs=dict(default=True, type="bool"),
            app_name_exact_match=dict(type="bool", default=False),
        ),
        required_one_of=[["app_name", "application_id"]],
        required_if=[("app_name_exact_match", True, ["app_name"])],
        supports_check_mode=True,
    )

    # build list of params
    params = {}
    if module.params["app_name"] and module.params["application_id"]:
        module.fail_json(msg="only one of 'app_name' or 'application_id' can be set")
    app_id = None
    if module.params["app_name"]:
        app_id = get_application_id(module)
    elif module.params["application_id"]:
        app_id = module.params["application_id"]
    else:
        module.fail_json(msg="you must set one of 'app_name' or 'application_id'")

    if app_id is None:
        module.fail_json(msg=f"No application with name {module.params['app_name']} is found in NewRelic")

    for item in ["changelog", "description", "revision", "user"]:
        if module.params[item]:
            params[item] = module.params[item]

    # If we're in check mode, just exit pretending like we succeeded
    if module.check_mode:
        module.exit_json(changed=True)

    # Send the data to New Relic
    url = f"https://api.newrelic.com/v2/applications/{quote(str(app_id), safe='')}/deployments.json"
    data = {"deployment": params}
    headers = {
        "Api-Key": module.params["token"],
        "Content-Type": "application/json",
    }
    response, info = fetch_url(module, url, data=module.jsonify(data), headers=headers, method="POST")
    if info["status"] in (200, 201):
        module.exit_json(changed=True)
    else:
        module.fail_json(msg=f"Unable to insert deployment marker: {info['msg']}")


def get_application_id(module):
    url = "https://api.newrelic.com/v2/applications.json"
    data = f"filter[name]={module.params['app_name']}"
    application_id = None
    headers = {
        "Api-Key": module.params["token"],
    }
    response, info = fetch_url(module, url, data=data, headers=headers)
    if info["status"] not in (200, 201):
        module.fail_json(msg=f"Unable to get application: {info['msg']}")

    result = json.loads(response.read())
    if result is None or len(result.get("applications", "")) == 0:
        module.fail_json(msg=f'No application found with name "{module.params["app_name"]}"')

    if module.params["app_name_exact_match"]:
        for item in result["applications"]:
            if item["name"] == module.params["app_name"]:
                application_id = item["id"]
                break
        if application_id is None:
            module.fail_json(msg=f'No application found with exact name "{module.params["app_name"]}"')
    else:
        application_id = result["applications"][0]["id"]

    return application_id


if __name__ == "__main__":
    main()



# Copyright (c) 2024, Stanislav Shamilov <shamilovstas@protonmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import re
import typing as t

from ansible_collections.community.general.plugins.module_utils._cmd_runner import CmdRunner, cmd_runner_fmt

if t.TYPE_CHECKING:
    from ansible.module_utils.basic import AnsibleModule


__state_map = {"present": "--install", "absent": "--uninstall"}

# sdkmanager --help 2>&1 | grep -A 2 -- --channel
__channel_map = {"stable": 0, "beta": 1, "dev": 2, "canary": 3}


def __map_channel(channel_name: str) -> int:
    if channel_name not in __channel_map:
        raise ValueError(f"Unknown channel name '{channel_name}'")
    return __channel_map[channel_name]


def sdkmanager_runner(module: AnsibleModule, **kwargs) -> CmdRunner:
    return CmdRunner(
        module,
        command="sdkmanager",
        arg_formats=dict(
            state=cmd_runner_fmt.as_map(__state_map),
            name=cmd_runner_fmt.as_list(),
            installed=cmd_runner_fmt.as_fixed("--list_installed"),
            list=cmd_runner_fmt.as_fixed("--list"),
            newer=cmd_runner_fmt.as_fixed("--newer"),
            sdk_root=cmd_runner_fmt.as_opt_eq_val("--sdk_root"),
            channel=cmd_runner_fmt.as_func(lambda x: [f"--channel={__map_channel(x)}"]),
        ),
        force_lang="C.UTF-8",  # Without this, sdkmanager binary crashes
        **kwargs,
    )


class Package:
    def __init__(self, name: str) -> None:
        self.name = name

    def __hash__(self) -> int:
        return hash(self.name)

    def __ne__(self, other: object) -> bool:
        if not isinstance(other, Package):
            return True
        return self.name != other.name

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Package):
            return False

        return self.name == other.name


class SdkManagerException(Exception):
    pass


class AndroidSdkManager:
    _RE_INSTALLED_PACKAGES_HEADER = re.compile(r"^Installed packages:$")
    _RE_UPDATABLE_PACKAGES_HEADER = re.compile(r"^Available Updates:$")

    # Example: '  platform-tools     | 27.0.0  | Android SDK Platform-Tools 27 | platform-tools  '
    _RE_INSTALLED_PACKAGE = re.compile(r"^\s*(?P<name>\S+)\s*\|\s*[0-9][^|]*\b\s*\|\s*.+\s*\|\s*(\S+)\s*$")

    # Example: '   platform-tools | 27.0.0    | 35.0.2'
    _RE_UPDATABLE_PACKAGE = re.compile(r"^\s*(?P<name>\S+)\s*\|\s*[0-9][^|]*\b\s*\|\s*[0-9].*\b\s*$")

    _RE_UNKNOWN_PACKAGE = re.compile(r"^Warning: Failed to find package \'(?P<package>\S+)\'\s*$")
    _RE_ACCEPT_LICENSE = re.compile(
        r"^The following packages can not be installed since their licenses or those of "
        r"the packages they depend on were not accepted"
    )

    def __init__(self, module: AnsibleModule) -> None:
        self.runner = sdkmanager_runner(module)

    def get_installed_packages(self) -> set[Package]:
        with self.runner("installed sdk_root channel") as ctx:
            rc, stdout, stderr = ctx.run()
            return self._parse_packages(stdout, self._RE_INSTALLED_PACKAGES_HEADER, self._RE_INSTALLED_PACKAGE)

    def get_updatable_packages(self) -> set[Package]:
        with self.runner("list newer sdk_root channel") as ctx:
            rc, stdout, stderr = ctx.run()
            return self._parse_packages(stdout, self._RE_UPDATABLE_PACKAGES_HEADER, self._RE_UPDATABLE_PACKAGE)

    def apply_packages_changes(self, packages: list[Package], accept_licenses: bool = False) -> tuple[int, str, str]:
        """Install or delete packages, depending on the `module.vars.state` parameter"""
        if len(packages) == 0:
            return 0, "", ""

        if accept_licenses:
            license_prompt_answer = "y"
        else:
            license_prompt_answer = "N"
        for package in packages:
            with self.runner("state name sdk_root channel", data=license_prompt_answer) as ctx:
                rc, stdout, stderr = ctx.run(name=package.name)

                for line in stdout.splitlines():
                    if self._RE_ACCEPT_LICENSE.match(line):
                        raise SdkManagerException("Licenses for some packages were not accepted")

                if rc != 0:
                    self._try_parse_stderr(stderr)
                    return rc, stdout, stderr
        return 0, "", ""

    def _try_parse_stderr(self, stderr: str) -> None:
        data = stderr.splitlines()
        for line in data:
            unknown_package_regex = self._RE_UNKNOWN_PACKAGE.match(line)
            if unknown_package_regex:
                package = unknown_package_regex.group("package")
                raise SdkManagerException(f"Unknown package {package}")

    @staticmethod
    def _parse_packages(stdout: str, header_regexp: re.Pattern, row_regexp: re.Pattern) -> set[Package]:
        data = stdout.splitlines()

        section_found: bool = False
        packages = set()

        for line in data:
            if not section_found:
                section_found = bool(header_regexp.match(line))
                continue
            else:
                p = row_regexp.match(line)
                if p:
                    packages.add(Package(p.group("name")))
        return packages



# -*- coding: utf-8 -*-

# (c) 2022 Red Hat Inc.
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

try:
    import boto3
    import botocore
except ImportError:
    pass  # will be captured by imported HAS_BOTO3


import typing
from typing import Any
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Union

if typing.TYPE_CHECKING:
    from typing_extensions import TypeAlias

    from .base import AWSPluginBase

    ClientType: TypeAlias = botocore.client.BaseClient
    ResourceType: TypeAlias = boto3.resources.base.ServiceResource
    BotoConn = Union[ClientType, ResourceType, Tuple[ClientType, ResourceType]]

from ansible.module_utils.basic import to_native

from ansible_collections.amazon.aws.plugins.module_utils.botocore import _aws_connection_info
from ansible_collections.amazon.aws.plugins.module_utils.botocore import _aws_region
from ansible_collections.amazon.aws.plugins.module_utils.botocore import _boto3_conn
from ansible_collections.amazon.aws.plugins.module_utils.exceptions import AnsibleBotocoreError


def boto3_conn(
    plugin: AWSPluginBase,
    conn_type: str,
    resource: Optional[str] = None,
    region: Optional[str] = None,
    endpoint: Optional[str] = None,
    **params,
) -> BotoConn:
    """
    Builds a boto3 resource/client connection cleanly wrapping the most common failures.
    Handles:
        ValueError, botocore.exceptions.BotoCoreError
    """

    try:
        return _boto3_conn(conn_type=conn_type, resource=resource, region=region, endpoint=endpoint, **params)
    except ValueError as e:
        plugin.fail_aws("Couldn't connect to AWS", exception=e)
    except botocore.exceptions.NoRegionError:
        plugin.fail_aws(
            f"The {plugin.ansible_name} plugin requires a region and none was found in configuration, "  # pyright: ignore[reportAttributeAccessIssue]
            "environment variables or module parameters"
        )
    except botocore.exceptions.BotoCoreError as e:
        plugin.fail_aws("Couldn't connect to AWS", exception=e)


def get_aws_connection_info(plugin: AWSPluginBase) -> Tuple[Optional[str], Optional[str], Dict[str, Any]]:
    try:
        return _aws_connection_info(plugin.get_options())  # pyright: ignore[reportAttributeAccessIssue]
    except AnsibleBotocoreError as e:
        plugin.fail_aws(to_native(e))


def get_aws_region(plugin: AWSPluginBase) -> Optional[str]:
    try:
        return _aws_region(plugin.get_options())  # pyright: ignore[reportAttributeAccessIssue]
    except AnsibleBotocoreError as e:
        plugin.fail_aws(to_native(e))



#!/usr/bin/python

# Copyright (c) 2012, Dag Wieers <dag@wieers.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: hponcfg
author: Dag Wieers (@dagwieers)
short_description: Configure HP iLO interface using C(hponcfg)
description:
  - This modules configures the HP iLO interface using C(hponcfg).
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: none
  diff_mode:
    support: none
options:
  path:
    description:
      - The XML file as accepted by C(hponcfg).
    required: true
    aliases: ['src']
    type: path
  minfw:
    description:
      - The minimum firmware level needed.
    type: str
  executable:
    description:
      - Path to the hponcfg executable (C(hponcfg) which uses E(PATH)).
    default: hponcfg
    type: str
  verbose:
    description:
      - Run C(hponcfg) in verbose mode (-v).
    default: false
    type: bool
requirements:
  - hponcfg tool
notes:
  - You need a working C(hponcfg) on the target system.
"""

EXAMPLES = r"""
- name: Example hponcfg configuration XML
  ansible.builtin.copy:
    content: |
      <ribcl VERSION="2.0">
        <login USER_LOGIN="user" PASSWORD="password">
          <rib_info MODE="WRITE">
            <mod_global_settings>
              <session_timeout value="0"/>
              <ssh_status value="Y"/>
              <ssh_port value="22"/>
              <serial_cli_status value="3"/>
              <serial_cli_speed value="5"/>
            </mod_global_settings>
          </rib_info>
        </login>
      </ribcl>
    dest: /tmp/enable-ssh.xml

- name: Configure HP iLO using enable-ssh.xml
  community.general.hponcfg:
    src: /tmp/enable-ssh.xml

- name: Configure HP iLO on VMware ESXi hypervisor
  community.general.hponcfg:
    src: /tmp/enable-ssh.xml
    executable: /opt/hp/tools/hponcfg
"""

from ansible_collections.community.general.plugins.module_utils._cmd_runner import CmdRunner, cmd_runner_fmt
from ansible_collections.community.general.plugins.module_utils._module_helper import ModuleHelper


class HPOnCfg(ModuleHelper):
    module = dict(
        argument_spec=dict(
            src=dict(type="path", required=True, aliases=["path"]),
            minfw=dict(type="str"),
            executable=dict(default="hponcfg", type="str"),
            verbose=dict(default=False, type="bool"),
        )
    )
    command_args_formats = dict(
        src=cmd_runner_fmt.as_opt_val("-f"),
        verbose=cmd_runner_fmt.as_bool("-v"),
        minfw=cmd_runner_fmt.as_opt_val("-m"),
    )

    def __run__(self):
        runner = CmdRunner(
            self.module,
            self.vars.executable,
            self.command_args_formats,
            check_rc=True,
        )
        runner(["src", "verbose", "minfw"]).run()

        # Consider every action a change (not idempotent yet!)
        self.changed = True


def main():
    HPOnCfg.execute()


if __name__ == "__main__":
    main()



# Copyright (c) 2021 Felix Fontein
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import traceback
import typing as t

from ansible.module_utils.common.text.converters import to_text

from ansible_collections.community.dns.plugins.module_utils._argspec import (
    ArgumentSpec,
    ModuleOptionProvider,
)
from ansible_collections.community.dns.plugins.module_utils._conversion.base import (
    DNSConversionError,
)
from ansible_collections.community.dns.plugins.module_utils._conversion.converter import (
    RecordConverter,
)
from ansible_collections.community.dns.plugins.module_utils._options import (
    create_record_transformation_argspec,
)
from ansible_collections.community.dns.plugins.module_utils._record import (
    format_record_for_output,
)
from ansible_collections.community.dns.plugins.module_utils._zone_record_api import (
    NOT_PROVIDED,
    DNSAPIAuthenticationError,
    DNSAPIError,
    ZoneRecordAPI,
)

from ._utils import get_prefix, normalize_dns_name

if t.TYPE_CHECKING:
    from ansible.module_utils.basic import AnsibleModule  # pragma: no cover

    from .._provider import ProviderInformation  # pragma: no cover
    from .._record import RecordIDT  # pragma: no cover
    from .._record_set import RecordSetIDT  # pragma: no cover
    from .._zone import ZoneIDT  # pragma: no cover
    from .._zone_record_api import NotProvidedType  # pragma: no cover
    from .._zone_record_set_api import ZoneRecordSetAPI  # pragma: no cover


def create_module_argument_spec(
    provider_information: ProviderInformation,
) -> ArgumentSpec:
    return ArgumentSpec(
        argument_spec={
            "what": {
                "type": "str",
                "choices": ["single_record", "all_types_for_record", "all_records"],
                "default": "single_record",
            },
            "zone_name": {"type": "str", "aliases": ["zone"]},
            "zone_id": {"type": provider_information.get_zone_id_type()},
            "record": {"type": "str"},
            "prefix": {"type": "str"},
            "type": {
                "type": "str",
                "choices": provider_information.get_supported_record_types(),
                "default": None,
            },
        },
        required_if=[
            ("what", "single_record", ["type"]),
            ("what", "single_record", ["record", "prefix"], True),
            ("what", "all_types_for_record", ["record", "prefix"], True),
        ],
        required_one_of=[
            ("zone_name", "zone_id"),
        ],
        mutually_exclusive=[
            ("zone_name", "zone_id"),
            ("record", "prefix"),
        ],
    ).merge(create_record_transformation_argspec())


def _run_module_record_api(
    module: AnsibleModule,
    provider_information: ProviderInformation,
    record_converter: RecordConverter,
    filter_record_type: str | NotProvidedType,
    filter_prefix: str | None | NotProvidedType,
    api: ZoneRecordAPI[ZoneIDT, RecordIDT],
) -> t.NoReturn:
    # Get zone information
    if module.params.get("zone_name") is not None:
        zone_in = normalize_dns_name(module.params.get("zone_name"))
        zone = api.get_zone_with_records_by_name(
            zone_in, prefix=filter_prefix, record_type=filter_record_type
        )
        if zone is None:
            module.fail_json(msg="Zone not found")
    else:
        zone = api.get_zone_with_records_by_id(
            module.params.get("zone_id"),
            prefix=filter_prefix,
            record_type=filter_record_type,
        )
        if zone is None:
            module.fail_json(msg="Zone not found")
        zone_in = normalize_dns_name(zone.zone.name)

    # Retrieve requested information
    records = []
    if module.params.get("what") in ("single_record", "all_types_for_record"):
        check_prefix = True
        record_in = normalize_dns_name(module.params.get("record"))
        prefix_in = module.params.get("prefix")
        record_in, prefix = get_prefix(
            normalized_zone=zone_in,
            normalized_record=record_in,
            prefix=prefix_in,
            provider_information=provider_information,
        )
    else:
        check_prefix = False
        prefix = None

    # Find matching records
    for record in zone.records:
        if check_prefix and record.prefix != prefix:
            continue
        records.append(
            (
                (record.prefix + "." + zone_in) if record.prefix else zone_in,
                record,
            )
        )

    # Convert records
    only_records = [record for record_name, record in records]
    record_converter.process_multiple_from_api(only_records)
    record_converter.process_multiple_to_user(only_records)

    # Format output
    data = [
        format_record_for_output(record, record_name, prefix=record.prefix)
        for record_name, record in records
    ]
    module.exit_json(
        changed=False,
        records=data,
        zone_id=zone.zone.id,
    )


def _run_module_record_set_api(
    module: AnsibleModule,
    provider_information: ProviderInformation,
    record_converter: RecordConverter,
    filter_record_type: str | NotProvidedType,
    filter_prefix: str | None | NotProvidedType,
    api: ZoneRecordSetAPI[ZoneIDT, RecordSetIDT, RecordIDT],
) -> t.NoReturn:
    # Get zone information
    if module.params.get("zone_name") is not None:
        zone_in = normalize_dns_name(module.params.get("zone_name"))
        zone = api.get_zone_with_record_sets_by_name(
            zone_in, prefix=filter_prefix, record_type=filter_record_type
        )
        if zone is None:
            module.fail_json(msg="Zone not found")
    else:
        zone = api.get_zone_with_record_sets_by_id(
            module.params.get("zone_id"),
            prefix=filter_prefix,
            record_type=filter_record_type,
        )
        if zone is None:
            module.fail_json(msg="Zone not found")
        zone_in = normalize_dns_name(zone.zone.name)

    # Retrieve requested information
    if module.params.get("what") in ("single_record", "all_types_for_record"):
        check_prefix = True
        record_in = normalize_dns_name(module.params.get("record"))
        prefix_in = module.params.get("prefix")
        record_in, prefix = get_prefix(
            normalized_zone=zone_in,
            normalized_record=record_in,
            prefix=prefix_in,
            provider_information=provider_information,
        )
    else:
        check_prefix = False
        prefix = None

    # Find matching records
    records = []
    for record_set in zone.record_sets:
        if check_prefix and record_set.prefix != prefix:
            continue
        records.extend(
            [
                (
                    (record.prefix + "." + zone_in) if record.prefix else zone_in,
                    record,
                )
                for record in record_set.records
            ]
        )

    # Convert records
    only_records = [record for record_name, record in records]
    record_converter.process_multiple_from_api(only_records)
    record_converter.process_multiple_to_user(only_records)

    # Format output
    data = [
        format_record_for_output(record, record_name, prefix=record.prefix)
        for record_name, record in records
    ]
    module.exit_json(
        changed=False,
        records=data,
        zone_id=zone.zone.id,
    )


def run_module(
    module: AnsibleModule,
    create_api: t.Callable[[], ZoneRecordAPI | ZoneRecordSetAPI],
    provider_information: ProviderInformation,
) -> t.NoReturn:
    option_provider = ModuleOptionProvider(module)
    record_converter = RecordConverter(provider_information, option_provider)
    record_converter.emit_deprecations(module.deprecate)

    filter_record_type: str | NotProvidedType = NOT_PROVIDED
    filter_prefix: str | None | NotProvidedType = NOT_PROVIDED
    if module.params.get("what") == "single_record":
        filter_record_type = module.params.get("type")
        if (
            filter_record_type
            and filter_record_type
            not in provider_information.get_supported_record_types()
        ):
            module.fail_json(msg=f"Invalid record type {filter_record_type}")
        if module.params.get("prefix") is not None:
            filter_prefix = provider_information.normalize_prefix(
                module.params.get("prefix")
            )
    elif module.params.get("what") == "all_types_for_record":
        if module.params.get("prefix") is not None:
            filter_prefix = provider_information.normalize_prefix(
                module.params.get("prefix")
            )

    try:
        # Create API
        api = create_api()

        if isinstance(api, ZoneRecordAPI):
            _run_module_record_api(
                module,
                provider_information,
                record_converter,
                filter_record_type,
                filter_prefix,
                api,
            )
        else:
            _run_module_record_set_api(
                module,
                provider_information,
                record_converter,
                filter_record_type,
                filter_prefix,
                api,
            )
    except DNSConversionError as e:
        module.fail_json(
            msg=f"Error while converting DNS values: {e.error_message}",
            error=e.error_message,
            exception=traceback.format_exc(),
        )
    except DNSAPIAuthenticationError as e:
        module.fail_json(
            msg=f"Cannot authenticate: {e}",
            error=to_text(e),
            exception=traceback.format_exc(),
        )
    except DNSAPIError as e:
        module.fail_json(
            msg=f"Error: {e}",
            error=to_text(e),
            exception=traceback.format_exc(),
        )



#
# -*- coding: utf-8 -*-
# Copyright 2020 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
#

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The eos_bgp_global config file.
It is in this file where the current configuration (as dict)
is compared to the provided configuration (as dict) and the command set
necessary to bring the current configuration to its desired end-state is
created.
"""

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.rm_base.resource_module import (
    ResourceModule,
)
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.utils import (
    dict_merge,
)

from ansible_collections.arista.eos.plugins.module_utils.network.eos.facts.facts import Facts
from ansible_collections.arista.eos.plugins.module_utils.network.eos.rm_templates.bgp_global import (
    Bgp_globalTemplate,
)


class Bgp_global(ResourceModule):
    """
    The eos_bgp_global config class
    """

    def __init__(self, module):
        super(Bgp_global, self).__init__(
            empty_fact_val={},
            facts_module=Facts(module),
            module=module,
            resource="bgp_global",
            tmplt=Bgp_globalTemplate(),
        )
        self.parsers = [
            "router",
            "vrf",
            "default_metric",
            "distance",
            "graceful_restart",
            "graceful_restart_helper",
            "maximum_paths",
            "monitoring",
            "route_target",
            "router_id",
            "shutdown",
            "timers",
            "ucmp_fec",
            "ucmp_link_bandwidth",
            "ucmp_mode",
            "update",
            "vlan",
            "vlan_aware_bundle",
        ]

    def execute_module(self):
        """Execute the module

        :rtype: A dictionary
        :returns: The result from module execution
        """
        if self.state not in ["parsed", "gathered"]:
            self.generate_commands()
            self.run_commands()
        return self.result

    def generate_commands(self):
        """Generate configuration commands to send based on
        want, have and desired state.
        """
        wantd = {}
        haved = {}
        if self.want.get("as_number") == self.have.get("as_number") or not self.have:
            if self.want:
                wantd = {self.want["as_number"]: self.want}
            if self.have:
                haved = {self.have["as_number"]: self.have}
        else:
            self._module.fail_json(
                msg="Only one bgp instance is allowed per device",
            )

        # turn all lists of dicts into dicts prior to merge
        for entry in wantd, haved:
            self._bgp_global_list_to_dict(entry)

        # if state is merged, merge want onto have and then compare
        if self.state == "merged":
            wantd = dict_merge(haved, wantd)

        # if state is deleted, empty out wantd and set haved to wantd
        if self.state in ["deleted", "purged"]:
            h_del = {}
            for k, v in haved.items():
                if k in wantd or not wantd:
                    h_del.update({k: v})
            wantd = {}
            haved = h_del

        if self.state == "deleted":
            self._compare(want={}, have=self.have)

        if self.state == "purged":
            for num, entry in haved.items():
                self.commands.append(
                    self._tmplt.render({"as_number": num}, "router", True),
                )

        for k, want in wantd.items():
            self._compare(want=want, have=haved.pop(k, {}))

    def _compare(self, want, have):
        """Leverages the base class `compare()` method and
        populates the list of commands to be run by comparing
        the `want` and `have` data with the `parsers` defined
        for the Bgp_global network resource.
        """
        self._compare_vrfs(want, have)
        self._compare_neighbor(want, have)
        self._compare_lists(want, have)
        self._compare_bgp_params(want, have)
        for name, entry in want.items():
            if name != "as_number":
                self.compare(
                    parsers=self.parsers,
                    want={name: entry},
                    have={name: have.pop(name, {})},
                )
        for name, entry in have.items():
            if name != "as_number":
                self.compare(
                    parsers=self.parsers,
                    want={},
                    have={name: have.get(name)},
                )

        if self.commands and "router bgp" not in self.commands[0]:
            self.commands.insert(
                0,
                self._tmplt.render(want or have, "router", False),
            )

    def _compare_bgp_params(self, want, have):
        parsers = [
            "bgp_params_additional_paths",
            "bgp_params_advertise_inactive",
            "bgp_params_allowas_in",
            "bgp_params_always_compare_med",
            "bgp_params_asn",
            "bgp_params_auto_local_addr",
            "bgp_params_bestpath_as_path",
            "bgp_params_bestpath_ecmp_fast",
            "bgp_params_bestpath_med",
            "bgp_params_bestpath_skip",
            "bgp_params_tie_break",
            "bgp_params_client_to_client",
            "bgp_params_cluster_id",
            "bgp_params_confederation",
            "bgp_params_control_plane_filter",
            "bgp_params_convergence",
            "bgp_params_default",
            "bgp_params.enforce_first_as",
            "bgp_params.host_routes",
            "bgp_params.labelled_unicast",
            "bgp_params.listen_limit",
            "bgp_params.listen_range",
            "bgp_params.log_neighbor_changes",
            "bgp_params.missing_policy",
            "bgp_params.monitoring",
            "bgp_params.nexthop_unchanged",
            "bgp_params.redistribute_internal",
            "bgp_params.route",
            "bgp_params.route_reflector",
            "bgp_params.transport",
        ]
        wbgp = want.pop("bgp_params", {})
        hbgp = have.pop("bgp_params", {})
        for name, entry in wbgp.items():
            if name == "bestpath":
                for k, v in entry.items():
                    h = {}
                    if hbgp.get(name):
                        h = {name: hbgp[name].pop(v, {})}
                    self.compare(
                        parsers=parsers,
                        want={"bgp_params": {name: {k: v}}},
                        have={"bgp_params": h},
                    )
                hbgp.pop(name, {})
                continue
            self.compare(
                parsers=parsers,
                want={"bgp_params": {name: entry}},
                have={"bgp_params": {name: hbgp.pop(name, {})}},
            )
        for name, entry in hbgp.items():
            self.compare(
                parsers=parsers,
                want={},
                have={"bgp_params": {name: entry}},
            )

    def _compare_vrfs(self, want, have):
        wvrf = want.pop("vrfs", {})
        hvrf = have.pop("vrfs", {})
        begin = len(self.commands)
        for name, entry in wvrf.items():
            self._compare_neighbor(entry, hvrf.get(name, {}))
            self._compare_lists(entry, hvrf.get(name, {}))
            self._compare_bgp_params(entry, hvrf.get(name, {}))
            for k, v in entry.items():
                if hvrf.get(name):
                    h = {k: hvrf[name].pop(k, {})}
                else:
                    h = {}
                if k != "vrf":
                    self.compare(parsers=self.parsers, want={k: v}, have=h)

            if len(self.commands) != begin:
                self.commands.insert(
                    begin,
                    self._tmplt.render({"vrf": name}, "vrf", False),
                )
                self.commands.append("exit")
        begin_negate = len(self.commands)
        for name, entry in hvrf.items():
            if name not in wvrf.keys():
                if self._check_af(name):
                    self._module.fail_json(
                        msg="Use the _bgp_address_family module to delete the address_family under vrf, before replacing/deleting the vrf.",
                    )
                else:
                    self.commands.append(
                        self._tmplt.render({"vrf": name}, "vrf", True),
                    )
                continue
            self.compare(parsers=self.parsers, want={}, have=entry)
            after_negate = len(self.commands)
            if after_negate != begin_negate:
                if "vrf " + name in self.commands:
                    index = self.commands.index("vrf " + name)
                    i = begin_negate
                    while i < after_negate:
                        cmd = self.commands.pop(i)
                        if cmd != "exit":
                            self.commands.insert(index + 1, cmd)
                        i += 1
                else:
                    self.commands.insert(
                        begin_negate,
                        self._tmplt.render({"vrf": name}, "vrf", False),
                    )
                    self.commands.append("exit")

    def _get_config(self, connection):
        return connection.get("show running-config | section bgp ")

    def _check_af(self, vrf):
        af_present = False
        if self._connection:
            config_lines = self._get_config(self._connection).splitlines()
            index = [i + 1 for i, el in enumerate(config_lines) if vrf in el]
            if index:
                # had to do this to escape flake8 and black errors
                ind = index[0]
                for line in config_lines[ind:]:
                    if "vrf" in line:
                        break
                    if "address-family" in line:
                        af_present = True
                        break
        return af_present

    def _compare_neighbor(self, want, have):
        parsers = [
            "neighbor.additional_paths",
            "neighbor.allowas_in",
            "neighbor.auto_local_addr",
            "neighbor.bfd",
            "neighbor.default_originate",
            "neighbor.description",
            "neighbor.dont_capability_negotiate",
            "neighbor.ebgp_multihop",
            "neighbor.encryption_password",
            "neighbor.enforce_first_as",
            "neighbor.export_localpref",
            "neighbor.fall_over",
            "neighbor.graceful_restart",
            "neighbor.graceful_restart_helper",
            "neighbor.idle_restart_timer",
            "neighbor.import_localpref",
            "neighbor.link_bandwidth",
            "neighbor.local_as",
            "neighbor.local_v6_addr",
            "neighbor.maximum_accepted_routes",
            "neighbor.maximum_received_routes",
            "neighbor.metric_out",
            "neighbor.monitoring",
            "neighbor.next_hop_self",
            "neighbor.next_hop_unchanged",
            "neighbor.next_hop_v6_addr",
            "neighbor.out_delay",
            "neighbor.remote_as",
            "neighbor.remove_private_as",
            "neighbor.peer_group",
            "neighbor.prefix_list",
            "neighbor.route_map",
            "neighbor.route_maps",
            "neighbor.route_reflector_client",
            "neighbor.route_to_peer",
            "neighbor.send_community",
            "neighbor.shutdown",
            "neighbor.soft_reconfiguration",
            "neighbor.transport",
            "neighbor.timers",
            "neighbor.ttl",
            "neighbor.update_source",
            "neighbor.weight",
        ]
        wneigh = want.pop("neighbor", {})
        hneigh = have.pop("neighbor", {})
        for name, entry in wneigh.items():
            peer = entry.get("peer") or entry["neighbor_address"]
            for k, v in entry.items():
                if k == "route_maps":
                    # Compare route_maps list: add each want item, remove each have item not in want
                    want_rms = v or []
                    have_rms = (hneigh.get(name) or {}).pop("route_maps", None) or []
                    want_set = {(rm["name"], rm["direction"]) for rm in want_rms}
                    have_set = {(rm["name"], rm["direction"]) for rm in have_rms}
                    for rm in want_rms:
                        if (rm["name"], rm["direction"]) not in have_set:
                            self.compare(
                                parsers=parsers,
                                want={"neighbor": {"neighbor_address": peer, "route_map": rm}},
                                have={"neighbor": {"neighbor_address": peer}},
                            )
                    for rm in have_rms:
                        if (rm["name"], rm["direction"]) not in want_set:
                            self.compare(
                                parsers=parsers,
                                want={},
                                have={"neighbor": {"neighbor_address": peer, "route_map": rm}},
                            )
                    continue
                if hneigh.get(name):
                    h = {"neighbor_address": peer, k: hneigh[name].pop(k, {})}
                else:
                    h = {}
                self.compare(
                    parsers=parsers,
                    want={"neighbor": {"neighbor_address": peer, k: v}},
                    have={"neighbor": h},
                )
        for name, entry in hneigh.items():
            if name not in wneigh.keys() and "peer_group" not in entry.keys():
                self.commands.append("no neighbor " + name)
                continue
            peer = entry.get("neighbor_address", name)
            for k, v in entry.items():
                if k == "route_maps":
                    # Already handled in want loop (we pop from have when processing want)
                    continue
                self.compare(
                    parsers=parsers,
                    want={},
                    have={"neighbor": {"neighbor_address": peer, k: v}},
                )

    def _compare_lists(self, want, have):
        for attrib in [
            "redistribute",
            "network",
            "aggregate_address",
            "access_group",
        ]:
            wdict = want.pop(attrib, {})
            hdict = have.pop(attrib, {})
            for key, entry in wdict.items():
                if entry != hdict.pop(key, {}):
                    self.addcmd(entry, attrib, False)
            # remove remaining items in have for replaced
            for entry in hdict.values():
                self.addcmd(entry, attrib, True)

    def _bgp_global_list_to_dict(self, entry):
        for name, proc in entry.items():
            if "neighbor" in proc:
                neigh_dict = {}
                for entry in proc.get("neighbor", []):
                    if entry.get("peer"):
                        peer = entry["peer"]
                    else:
                        peer = entry["neighbor_address"]
                    # Normalize: convert deprecated route_map (single) to route_maps list
                    if "route_map" in entry and entry.get("route_map"):
                        if "route_maps" not in entry:
                            entry["route_maps"] = [entry["route_map"]]
                        entry.pop("route_map", None)
                    neigh_dict.update({peer: entry})
                proc["neighbor"] = neigh_dict

            if "network" in proc:
                network_dict = {}
                for entry in proc.get("network", []):
                    network_dict.update({entry["address"]: entry})
                proc["network"] = network_dict

            if "aggregate_address" in proc:
                agg_dict = {}
                for entry in proc.get("aggregate_address", []):
                    agg_dict.update({entry["address"]: entry})
                proc["aggregate_address"] = agg_dict

            if "access_group" in proc:
                access_dict = {}
                for entry in proc.get("access_group", []):
                    access_dict.update({entry["afi"]: entry})
                proc["access_group"] = access_dict

            if "redistribute" in proc:
                redis_dict = {}
                for entry in proc.get("redistribute", []):
                    redis_dict.update({entry["protocol"]: entry})
                proc["redistribute"] = redis_dict

            if "vrfs" in proc:
                vrf_dict = {}
                for entry in proc.get("vrfs", []):
                    vrf_dict.update({entry["vrf"]: entry})
                proc["vrfs"] = vrf_dict
                self._bgp_global_list_to_dict(proc["vrfs"])



# -*- coding: utf-8 -*-
# Copyright 2021 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

"""
Test plugin file for netaddr tests: ipv4_hostmask
"""

from __future__ import absolute_import, division, print_function

from ansible_collections.ansible.utils.plugins.plugin_utils.base.ipaddress_utils import (
    _need_ipaddress,
    ip_network,
)
from ansible_collections.ansible.utils.plugins.plugin_utils.base.utils import _validate_args


__metaclass__ = type

DOCUMENTATION = """
    name: ipv4_hostmask
    author: Priyam Sahoo (@priyamsahoo)
    version_added: "2.2.0"
    short_description: Test if an address is a valid hostmask
    description:
        - This plugin checks if the provided ip address is a IPv4 hostmask or not
    options:
        ip:
            description:
            - A string that represents the value against which the test is going to be performed
            - 'For example: C(0.1.255.255) or C(255.255.255.0)'
            type: str
            required: True
    notes:
"""

EXAMPLES = r"""

#### Simple examples

- name: Check if 0.0.0.255 is a hostmask
  ansible.builtin.set_fact:
    data: "{{ '0.0.0.255' is ansible.utils.ipv4_hostmask }}"

# TASK [Check if 0.0.0.255 is a hostmask] ***********************************************
# ok: [localhost] => {
#     "ansible_facts": {
#         "data": true
#     },
#     "changed": false
# }

- name: Check if 255.255.255.0 is not a hostmask
  ansible.builtin.set_fact:
    data: "{{ '255.255.255.0' is not ansible.utils.ipv4_hostmask }}"

# TASK [Check if 255.255.255.0 is a hostmask] *********************************
# ok: [localhost] => {
#     "ansible_facts": {
#         "data": true
#     },
#     "changed": false
# }
"""

RETURN = """
  data:
    description:
      - If jinja test satisfies plugin expression C(true)
      - If jinja test does not satisfy plugin expression C(false)
"""


@_need_ipaddress
def _ipv4_hostmask(ip):
    """Test if an address is a hostmask"""

    params = {"ip": ip}
    _validate_args("ipv4_hostmask", DOCUMENTATION, params)

    try:
        ipaddr = ip_network("10.0.0.0/{ip}".format(ip=ip))
        return str(ipaddr.hostmask) == ip
    except Exception:
        return False


class TestModule(object):
    """network jinja test"""

    test_map = {"ipv4_hostmask": _ipv4_hostmask}

    def tests(self):
        return self.test_map
